# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import webtest

from pyramid.response import Response
from pyramid.view import view_config
from pyramid.config import Configurator
import pyramid.httpexceptions as exc


@view_config(route_name="home")
def home_view(request):
    return Response("<p>INDEX RESPONSE</p>")


@view_config(route_name="error")
def error(request):
    raise RuntimeError("error")


@view_config(route_name="not_found_exception_response")
def not_found_exception_response(request):
    raise exc.exception_response(404)


@view_config(route_name="raise_not_found")
def raise_not_found(request):
    raise exc.HTTPNotFound()


@view_config(route_name="return_not_found")
def return_not_found(request):
    return exc.HTTPNotFound()


@view_config(route_name="redirect")
def redirect(request):
    raise exc.HTTPFound(request.route_url("home"))


@view_config(route_name="html_insertion")
def html_insertion(request):
    return Response(
        "<!DOCTYPE html><html><head>Some header</head>"
        "<body><h1>My First Heading</h1><p>My first paragraph.</p>"
        "</body></html>"
    )


class RestView:
    def __init__(self, request):
        self.request = request

    @view_config(route_name="rest", request_method="GET")
    def get(self):
        return Response("Called GET")

    @view_config(route_name="rest", request_method="POST")
    def post(self):
        return Response("Called POST")


def simple_tween_factory(handler, registry):
    def simple_tween(request):
        return handler(request)

    return simple_tween


# CORNICE SECTION
try:
    import cornice
    import cornice.resource
except ImportError:
    cornice = None
else:
    service = cornice.Service(name="service", path="/service", description="Service")


    @service.get()
    def cornice_service_get_info(request):
        return {"Hello": "World"}


    @cornice.resource.resource(collection_path="/resource", path="/resource/{id}")
    class Resource(object):
        def __init__(self, request, context=None):
            self.request = request

        def collection_get(self):
            return {"resource": ["1", "2"]}

        @cornice.resource.view(renderer="json")
        def get(self):
            return self.request.matchdict["id"]


    cornice_error = cornice.Service(
        name="cornice_error", path="/cornice_error", description="Error"
    )


    @cornice_error.get()
    def cornice_error_get_info(request):
        raise RuntimeError("error")


# END CORNICE SECTION


def target_application(with_tweens=False, tweens_explicit=False, handle_exceptions=False):
    settings = {}
    if with_tweens and tweens_explicit:
        settings["pyramid.tweens"] = [
            "_test_application:simple_tween_factory",
        ]

    config = Configurator(settings=settings)
    if cornice:

        # Use an execution policy which doesn't re-trigger the error handler
        def execution_policy(environ, router):
            with router.request_context(environ) as request:
                return router.invoke_request(request)

        config.set_execution_policy(execution_policy)
        config.add_settings(handle_exceptions=handle_exceptions)
        config.include("cornice")

    config.add_route("home", "/")
    config.add_route("html_insertion", "/html_insertion")
    config.add_route("error", "/error")
    config.add_route("not_found_exception_response", "/nf1")
    config.add_route("raise_not_found", "/nf2")
    config.add_route("return_not_found", "/nf3")
    config.add_route("redirect", "/redirect")
    config.add_route("rest", "/rest")
    if with_tweens:
        config.add_tween("_test_application:simple_tween_factory")
    config.scan()

    application = config.make_wsgi_app()

    _test_application = webtest.TestApp(application)

    return _test_application



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from newrelic.api.html_insertion import insert_html_snippet
from newrelic.api.transaction import current_transaction
from newrelic.common.object_wrapper import wrap_function_wrapper
from newrelic.config import extra_settings
from newrelic.packages import six

_logger = logging.getLogger(__name__)

_boolean_states = {
    "1": True,
    "yes": True,
    "true": True,
    "on": True,
    "0": False,
    "no": False,
    "false": False,
    "off": False,
}


def _setting_boolean(value):
    if value.lower() not in _boolean_states:
        raise ValueError("Not a boolean: %s" % value)
    return _boolean_states[value.lower()]


_settings_types = {
    "browser_monitoring.auto_instrument": _setting_boolean,
    "browser_monitoring.auto_instrument_passthrough": _setting_boolean,
}

_settings_defaults = {
    "browser_monitoring.auto_instrument": True,
    "browser_monitoring.auto_instrument_passthrough": True,
}

flask_compress_settings = extra_settings(
    "import-hook:flask_compress", types=_settings_types, defaults=_settings_defaults
)


def _nr_wrapper_Compress_after_request(wrapped, instance, args, kwargs):
    def _params(response, *args, **kwargs):
        return response

    response = _params(*args, **kwargs)

    # Need to be running within a valid web transaction.

    transaction = current_transaction()

    if not transaction:
        return wrapped(*args, **kwargs)

    # Only insert RUM JavaScript headers if enabled
    # in configuration and not already likely inserted.

    if not transaction.settings.browser_monitoring.enabled:
        return wrapped(*args, **kwargs)

    if transaction.autorum_disabled:
        return wrapped(*args, **kwargs)

    if not flask_compress_settings.browser_monitoring.auto_instrument:
        return wrapped(*args, **kwargs)

    if transaction.rum_header_generated:
        return wrapped(*args, **kwargs)

    # Only possible if the content type is one of the allowed
    # values. Normally this is just text/html, but optionally
    # could be defined to be list of further types. For example
    # a user may want to also perform insertion for
    # 'application/xhtml+xml'.

    ctype = (response.mimetype or "").lower()

    if ctype not in transaction.settings.browser_monitoring.content_type:
        return wrapped(*args, **kwargs)

    # Don't risk it if content encoding already set.

    if "Content-Encoding" in response.headers:
        return wrapped(*args, **kwargs)

    # Don't risk it if content is actually within an attachment.

    cdisposition = response.headers.get("Content-Disposition", "").lower()

    if cdisposition.split(";")[0].strip() == "attachment":
        return wrapped(*args, **kwargs)

    # No point continuing if header is empty. This can occur if RUM is not enabled within the UI. We don't want to
    # generate the header just yet as we want to do that as late as possible so that application server time in header
    # is as accurate as possible. In particular, if the response content is generated on demand then the flattening
    # of the response could take some time and we want to track that. We thus generate header below at
    # the point of insertion.

    # If the response has direct_passthrough flagged, then is
    # likely to be streaming a file or other large response.
    direct_passthrough = getattr(response, "direct_passthrough", None)
    if direct_passthrough:
        if not (flask_compress_settings.browser_monitoring.auto_instrument_passthrough):
            return wrapped(*args, **kwargs)

        # In those cases, if the mimetype is still a supported browser
        # insertion mimetype is not an attachment, and will be compressed, then
        # we should try to go ahead and insert browser stuff since Flask
        # Compress change the response anyway.
        #
        # In order to do that, we have to disable direct_passthrough on the
        # response since we have to immediately read the contents of the file.
        elif ctype == "text/html":
            response.direct_passthrough = False
        else:
            return wrapped(*args, **kwargs)

    # Make sure we flatten any content first as it could be
    # stored as a list of strings in the response object. We
    # assign it back to the response object to avoid having
    # multiple copies of the string in memory at the same time
    # as we progress through steps below.

    result = insert_html_snippet(response.get_data(), lambda: six.b(transaction.browser_timing_header()))

    if result is not None:
        if transaction.settings.debug.log_autorum_middleware:
            _logger.debug(
                "RUM insertion from flask_compress " "triggered. Bytes added was %r.",
                len(result) - len(response.get_data()),
            )

        response.set_data(result)
        response.headers["Content-Length"] = str(len(response.get_data()))

    return wrapped(*args, **kwargs)


def instrument_flask_compress(module):
    wrap_function_wrapper(module, "Compress.after_request", _nr_wrapper_Compress_after_request)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.common.encoding_utils import unpack_field
from newrelic.common.object_wrapper import transient_function_wrapper
from newrelic.core.database_utils import SQLConnections


def validate_tt_parameters(required_params=None, forgone_params=None):
    required_params = required_params or {}
    forgone_params = forgone_params or {}

    @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.record_transaction")
    def _validate_tt_parameters(wrapped, instance, args, kwargs):
        try:
            result = wrapped(*args, **kwargs)
        except:
            raise

        # Now that transaction has been recorded, generate
        # a transaction trace

        connections = SQLConnections()
        trace_data = instance.transaction_trace_data(connections)
        pack_data = unpack_field(trace_data[0][4])
        tt_intrinsics = pack_data[0][4]["intrinsics"]

        for name in required_params:
            assert name in tt_intrinsics, "name=%r, intrinsics=%r" % (name, tt_intrinsics)
            assert tt_intrinsics[name] == required_params[name], "name=%r, value=%r, intrinsics=%r" % (
                name,
                required_params[name],
                tt_intrinsics,
            )

        for name in forgone_params:
            assert name not in tt_intrinsics, "name=%r, intrinsics=%r" % (name, tt_intrinsics)

        return result

    return _validate_tt_parameters



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from testing_support.fixtures import (
    override_application_settings,
    validate_tt_parenting,
)
from testing_support.util import instance_hostname
from testing_support.validators.validate_tt_collector_json import (
    validate_tt_collector_json,
)

from newrelic.api.background_task import background_task

from conftest import ES_SETTINGS, ES_VERSION

# Settings

_enable_instance_settings = {
    "datastore_tracer.instance_reporting.enabled": True,
    "datastore_tracer.database_name_reporting.enabled": True,
}
_disable_instance_settings = {
    "datastore_tracer.instance_reporting.enabled": False,
    "datastore_tracer.database_name_reporting.enabled": False,
}
_instance_only_settings = {
    "datastore_tracer.instance_reporting.enabled": True,
    "datastore_tracer.database_name_reporting.enabled": False,
}

# Expected parameters

_enabled_required = {
    "host": instance_hostname(ES_SETTINGS["host"]),
    "port_path_or_id": str(ES_SETTINGS["port"]),
}
_enabled_forgone = {
    "db.instance": "VALUE NOT USED",
}

_disabled_required = {}
_disabled_forgone = {
    "host": "VALUE NOT USED",
    "port_path_or_id": "VALUE NOT USED",
    "db.instance": "VALUE NOT USED",
}

_instance_only_required = {
    "host": instance_hostname(ES_SETTINGS["host"]),
    "port_path_or_id": str(ES_SETTINGS["port"]),
}
_instance_only_forgone = {
    "db.instance": "VALUE NOT USED",
}

_tt_parenting = (
    "TransactionNode",
    [
        ("DatastoreNode", []),
    ],
)


# Query


def _exercise_es_v7(es):
    es.index(index="contacts", doc_type="person", body={"name": "Joe Tester", "age": 25, "title": "QA Master"}, id=1)


def _exercise_es_v8(es):
    es.index(index="contacts", body={"name": "Joe Tester", "age": 25, "title": "QA Master"}, id=1)


_exercise_es = _exercise_es_v7 if ES_VERSION < (8, 0, 0) else _exercise_es_v8

# Tests


@override_application_settings(_enable_instance_settings)
@validate_tt_collector_json(datastore_params=_enabled_required, datastore_forgone_params=_enabled_forgone)
@validate_tt_parenting(_tt_parenting)
@background_task()
def test_trace_node_datastore_params_enable_instance(client):
    _exercise_es(client)


@override_application_settings(_disable_instance_settings)
@validate_tt_collector_json(datastore_params=_disabled_required, datastore_forgone_params=_disabled_forgone)
@validate_tt_parenting(_tt_parenting)
@background_task()
def test_trace_node_datastore_params_disable_instance(client):
    _exercise_es(client)


@override_application_settings(_instance_only_settings)
@validate_tt_collector_json(datastore_params=_instance_only_required, datastore_forgone_params=_instance_only_forgone)
@validate_tt_parenting(_tt_parenting)
@background_task()
def test_trace_node_datastore_params_instance_only(client):
    _exercise_es(client)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from conftest import logger as conf_logger
import logging
import pytest

from newrelic.api.background_task import background_task
from newrelic.api.log import NewRelicLogForwardingHandler
from testing_support.fixtures import reset_core_stats_engine
from testing_support.validators.validate_log_event_count import validate_log_event_count
from testing_support.validators.validate_log_event_count_outside_transaction import validate_log_event_count_outside_transaction
from testing_support.validators.validate_function_called import validate_function_called




@pytest.fixture(scope="function")
def uninstrument_logging():
    instrumented = logging.Logger.callHandlers
    while hasattr(logging.Logger.callHandlers, "__wrapped__"):
        logging.Logger.callHandlers = logging.Logger.callHandlers.__wrapped__
    yield
    logging.Logger.callHandlers = instrumented


@pytest.fixture(scope="function")
def logger(conf_logger, uninstrument_logging):
    handler = NewRelicLogForwardingHandler()
    conf_logger.addHandler(handler)
    yield conf_logger
    conf_logger.removeHandler(handler)


def exercise_logging(logger):
    logger.warning("C")
    assert len(logger.caplog.records) == 1


def test_handler_inside_transaction(logger):
    @validate_log_event_count(1)
    @validate_function_called("newrelic.api.log", "NewRelicLogForwardingHandler.emit")
    @background_task()
    def test():
        exercise_logging(logger)

    test()


@reset_core_stats_engine()
def test_handler_outside_transaction(logger):
    @validate_log_event_count_outside_transaction(1)
    @validate_function_called("newrelic.api.log", "NewRelicLogForwardingHandler.emit")
    def test():
        exercise_logging(logger)

    test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import deque

import pytest
import sanic
from testing_support.fixtures import (
    function_not_called,
    override_application_settings,
    override_generic_settings,
    override_ignore_status_codes,
)
from testing_support.validators.validate_code_level_metrics import (
    validate_code_level_metrics,
)
from testing_support.validators.validate_transaction_errors import (
    validate_transaction_errors,
)
from testing_support.validators.validate_transaction_event_attributes import (
    validate_transaction_event_attributes,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.application import application_instance
from newrelic.api.external_trace import ExternalTrace
from newrelic.api.transaction import Transaction
from newrelic.common.package_version_utils import get_package_version
from newrelic.core.config import global_settings

SANIC_VERSION = tuple(map(int, get_package_version("sanic").split(".")))

sanic_21 = SANIC_VERSION >= (21,)
sanic_v19_to_v22_12 = SANIC_VERSION >= (19,) and SANIC_VERSION < (22, 12)

BASE_METRICS = [
    ("Function/_target_application:index", 1),
    ("Function/_target_application:request_middleware", 1 if sanic_v19_to_v22_12 else 2),
]
FRAMEWORK_METRICS = [
    ("Python/Framework/Sanic/%s" % sanic.__version__, 1),
]
BASE_ATTRS = ["response.status", "response.headers.contentType", "response.headers.contentLength"]

validate_base_transaction_event_attr = validate_transaction_event_attributes(
    required_params={"agent": BASE_ATTRS, "user": [], "intrinsic": []},
)


@validate_code_level_metrics("_target_application", "index")
@validate_transaction_metrics(
    "_target_application:index",
    scoped_metrics=BASE_METRICS,
    rollup_metrics=BASE_METRICS + FRAMEWORK_METRICS,
)
@validate_base_transaction_event_attr
def test_simple_request(app):
    response = app.fetch("get", "/")
    assert response.status == 200


@function_not_called("newrelic.core.stats_engine", "StatsEngine.record_transaction")
def test_websocket(app):
    headers = {"upgrade": "WebSocket"}
    response = app.fetch("get", "/", headers=headers)
    assert response.status == 200


@pytest.mark.parametrize(
    "method",
    (
        "get",
        "post",
        "put",
        "patch",
        "delete",
    ),
)
def test_method_view(app, method):
    metric_name = "Function/_target_application:MethodView." + method

    @validate_code_level_metrics("_target_application.MethodView", method)
    @validate_transaction_metrics(
        "_target_application:MethodView." + method,
        scoped_metrics=[(metric_name, 1)],
        rollup_metrics=[(metric_name, 1)],
    )
    @validate_base_transaction_event_attr
    def _test():
        response = app.fetch(method, "/method_view")
        assert response.status == 200

    _test()


DT_METRICS = [
    ("Supportability/DistributedTrace/AcceptPayload/Success", None),
    ("Supportability/TraceContext/TraceParent/Accept/Success", 1),
]


@validate_transaction_metrics(
    "_target_application:index",
    scoped_metrics=BASE_METRICS,
    rollup_metrics=BASE_METRICS + DT_METRICS + FRAMEWORK_METRICS,
)
@validate_base_transaction_event_attr
@override_application_settings(
    {
        "distributed_tracing.enabled": True,
    }
)
def test_inbound_distributed_trace(app):
    transaction = Transaction(application_instance())
    dt_headers = ExternalTrace.generate_request_headers(transaction)

    response = app.fetch("get", "/", headers=dict(dt_headers))
    assert response.status == 200


@pytest.mark.parametrize("endpoint", ["error", "write_response_error"])
def test_recorded_error(app, endpoint, sanic_version):
    if sanic_version >= (21, 0, 0) and endpoint == "write_response_error":
        pytest.skip()

    ERROR_METRICS = [
        ("Function/_target_application:%s" % endpoint, 1),
    ]

    @validate_transaction_errors(errors=["builtins:ValueError"])
    @validate_base_transaction_event_attr
    @validate_transaction_metrics(
        "_target_application:%s" % endpoint,
        scoped_metrics=ERROR_METRICS,
        rollup_metrics=ERROR_METRICS + FRAMEWORK_METRICS,
    )
    def _test():
        if endpoint == "write_response_error":
            with pytest.raises(ValueError):
                response = app.fetch("get", "/" + endpoint)
        else:
            response = app.fetch("get", "/" + endpoint)
            assert response.status == 500

    _test()


NOT_FOUND_METRICS = [
    ("Function/_target_application:not_found", 1),
]


@validate_transaction_metrics(
    "_target_application:not_found",
    scoped_metrics=NOT_FOUND_METRICS,
    rollup_metrics=NOT_FOUND_METRICS + FRAMEWORK_METRICS,
)
@validate_base_transaction_event_attr
@override_ignore_status_codes([404])
@validate_transaction_errors(errors=[])
def test_ignored_by_status_error(app):
    response = app.fetch("get", "/404")
    assert response.status == 404


DOUBLE_ERROR_METRICS = [
    ("Function/_target_application:zero_division_error", 1),
]


@validate_transaction_metrics(
    "_target_application:zero_division_error",
    scoped_metrics=DOUBLE_ERROR_METRICS,
    rollup_metrics=DOUBLE_ERROR_METRICS,
)
@validate_transaction_errors(errors=["builtins:ValueError", "builtins:ZeroDivisionError"])
def test_error_raised_in_error_handler(app):
    # Because of a bug in Sanic versions <0.8.0, the response.status value is
    # inconsistent. Rather than assert the status value, we rely on the
    # transaction errors validator to confirm the application acted as we'd
    # expect it to.
    app.fetch("get", "/zero")


STREAMING_ATTRS = ["response.status", "response.headers.contentType"]
STREAMING_METRICS = [
    ("Function/_target_application:streaming", 1),
]


@validate_code_level_metrics("_target_application", "streaming")
@validate_transaction_metrics(
    "_target_application:streaming",
    scoped_metrics=STREAMING_METRICS,
    rollup_metrics=STREAMING_METRICS,
)
@validate_transaction_event_attributes(
    required_params={"agent": STREAMING_ATTRS, "user": [], "intrinsic": []},
)
def test_streaming_response(app):
    # streaming responses do not have content-length headers
    response = app.fetch("get", "/streaming")
    assert response.status == 200


ERROR_IN_ERROR_TESTS = [
    (
        "/sync-error",
        "_target_application:sync_error",
        [
            ("Function/_target_application:sync_error", 1),
            ("Function/_target_application:handle_custom_exception_sync", 1),
        ],
        ["_target_application:CustomExceptionSync", "sanic.exceptions:SanicException"],
    ),
    (
        "/async-error",
        "_target_application:async_error",
        [
            ("Function/_target_application:async_error", 1),
            ("Function/_target_application:handle_custom_exception_async", 1),
        ],
        ["_target_application:CustomExceptionAsync"],
    ),
]


@pytest.mark.parametrize("url,metric_name,metrics,errors", ERROR_IN_ERROR_TESTS)
@pytest.mark.parametrize("nr_enabled", (True, False))
def test_errors_in_error_handlers(nr_enabled, app, url, metric_name, metrics, errors):
    settings = global_settings()

    @override_generic_settings(settings, {"enabled": nr_enabled})
    def _test():
        # Because of a bug in Sanic versions <0.8.0, the response.status value
        # is inconsistent. Rather than assert the status value, we rely on the
        # transaction errors validator to confirm the application acted as we'd
        # expect it to.
        app.fetch("get", url)

    if nr_enabled:
        _test = validate_transaction_errors(errors=errors)(_test)
        _test = validate_transaction_metrics(metric_name, scoped_metrics=metrics, rollup_metrics=metrics)(_test)
    else:
        _test = function_not_called("newrelic.core.stats_engine", "StatsEngine.record_transaction")(_test)

    _test()


def test_no_transaction_when_nr_disabled(app):
    settings = global_settings()

    @function_not_called("newrelic.core.stats_engine", "StatsEngine.record_transaction")
    @override_generic_settings(settings, {"enabled": False})
    def _test():
        app.fetch("GET", "/")

    _test()


async def async_returning_middleware(*args, **kwargs):
    from sanic.response import json

    return json({"oops": "I returned it again"})


def sync_returning_middleware(*args, **kwargs):
    from sanic.response import json

    return json({"oops": "I returned it again"})


def sync_failing_middleware(*args, **kwargs):
    from sanic.exceptions import SanicException

    raise SanicException("Everything is ok", status_code=200)


@pytest.mark.parametrize(
    "middleware,attach_to,metric_name,transaction_name",
    [
        (
            async_returning_middleware,
            "request",
            "test_application:async_returning_middleware",
            "test_application:async_returning_middleware",
        ),
        (
            sync_returning_middleware,
            "request",
            "test_application:sync_returning_middleware",
            "test_application:sync_returning_middleware",
        ),
        (
            sync_failing_middleware,
            "request",
            "test_application:sync_failing_middleware",
            "test_application:sync_failing_middleware",
        ),
        (
            async_returning_middleware,
            "response",
            "test_application:async_returning_middleware",
            "_target_application:index",
        ),
        (
            sync_returning_middleware,
            "response",
            "test_application:sync_returning_middleware",
            "_target_application:index",
        ),
    ],
)
def test_returning_middleware(app, middleware, attach_to, metric_name, transaction_name):

    metrics = [
        ("Function/%s" % metric_name, 1),
    ]

    @validate_code_level_metrics(*metric_name.split(":"))
    @validate_transaction_metrics(
        transaction_name,
        scoped_metrics=metrics,
        rollup_metrics=metrics,
    )
    @validate_base_transaction_event_attr
    def _test():
        response = app.fetch("get", "/")
        assert response.status == 200

    original_request_middleware = deque(app.app.request_middleware)
    original_response_middleware = deque(app.app.response_middleware)
    app.app.register_middleware(middleware, attach_to)

    try:
        _test()
    finally:
        app.app.request_middleware = original_request_middleware
        app.app.response_middleware = original_response_middleware


def error_middleware(*args, **kwargs):
    raise ValueError("1 != 0")


def test_errors_in_middleware(app):
    metrics = [("Function/test_application:error_middleware", 1)]

    @validate_transaction_metrics(
        "test_application:error_middleware",
        scoped_metrics=metrics,
        rollup_metrics=metrics,
    )
    @validate_base_transaction_event_attr
    @validate_transaction_errors(errors=["builtins:ValueError"])
    def _test():
        response = app.fetch("get", "/")
        assert response.status == 500

    original_request_middleware = deque(app.app.request_middleware)
    original_response_middleware = deque(app.app.response_middleware)
    app.app.register_middleware(error_middleware, "request")

    try:
        _test()
    finally:
        app.app.request_middleware = original_request_middleware
        app.app.response_middleware = original_response_middleware


BLUEPRINT_METRICS = [
    ("Function/_target_application:blueprint_middleware", 1),
    ("Function/_target_application:blueprint_route", 1),
]


@validate_code_level_metrics("_target_application", "blueprint_middleware")
@validate_code_level_metrics("_target_application", "blueprint_route")
@validate_transaction_metrics(
    "_target_application:blueprint_route",
    scoped_metrics=BLUEPRINT_METRICS,
    rollup_metrics=BLUEPRINT_METRICS + FRAMEWORK_METRICS,
)
@validate_transaction_errors(errors=[])
def test_blueprint_middleware(app):
    response = app.fetch("get", "/blueprint")
    assert response.status == 200


def test_unknown_route(app, sanic_version):
    _tx_name = (
        "_target_application:CustomRouter.get" if sanic_version[0] < 21 else "_target_application:request_middleware"
    )

    @validate_transaction_metrics(_tx_name)
    def _test():
        response = app.fetch("get", "/what-route")
        assert response.status == 404

    _test()


def test_bad_method(app, sanic_version):
    _tx_name = (
        "_target_application:CustomRouter.get" if sanic_version[0] < 21 else "_target_application:request_middleware"
    )

    @validate_transaction_metrics(_tx_name)
    @override_ignore_status_codes([405])
    @validate_transaction_errors(errors=[])
    def _test():
        response = app.fetch("post", "/")
        assert response.status == 405

    _test()


@pytest.fixture
def sanic_version():
    return tuple(int(v) for v in sanic.__version__.split("."))



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import functools
import time

from testing_support.fixtures import override_generic_settings
from newrelic.core.config import global_settings
from newrelic.core.agent_protocol import AgentProtocol
from newrelic.common.agent_http import DeveloperModeClient
from newrelic.common.encoding_utils import json_encode


DEFAULT = object()
LINKED_APPLICATIONS = []
ENVIRONMENT = []
NOW = time.time()
EMPTY_SAMPLES = {
    'reservoir_size': 100,
    'events_seen': 0,
}


_all_endpoints = (
    ('send_metric_data', (NOW, NOW + 1, ())),
    ('send_transaction_events', (EMPTY_SAMPLES, ())),
    ('send_custom_events', (EMPTY_SAMPLES, ())),
    ('send_error_events', (EMPTY_SAMPLES, ())),
    ('send_transaction_traces', ([[]],)),
    ('send_sql_traces', ([[]],)),
    ('get_agent_commands', ()),
    ('send_profile_data', ([[]],)),
    ('send_errors', ([[]],)),
    ('send_agent_command_results', ({0: {}},)),
    ('agent_settings', ({},)),
    ('send_span_events', (EMPTY_SAMPLES, ())),
    ('shutdown_session', ()),
)


class CustomTestClient(DeveloperModeClient):
    def __init__(self, connect_response_fields, *args, **kwargs):
        super(CustomTestClient, self).__init__(*args, **kwargs)
        self.connect_response_fields = connect_response_fields
        self.headers = []

    def log_request(self, fp, method, url, params, payload, headers):
        self.headers.append(headers)
        return super(CustomTestClient, self).log_request(fp, method, url, params, payload, headers)

    def send_request(
        self,
        method="POST",
        path="/agent_listener/invoke_raw_method",
        params=None,
        headers=None,
        payload=None,
    ):
        agent_method = params["method"]
        if agent_method == "connect" and self.connect_response_fields is not DEFAULT:
            connect_response = dict(self.RESPONSES[agent_method])
            connect_response.update(self.connect_response_fields)
            payload = {"return_value": connect_response}
            response_data = json_encode(payload).encode("utf-8")
            return (200, response_data)
        else:
            return super(CustomTestClient, self).send_request(
                method,
                path,
                params,
                headers,
                payload,
            )


@pytest.mark.parametrize('headers_map_present', (True, False))
def test_no_blob_behavior(headers_map_present):
    if headers_map_present:
        connect_response_fields = {u"request_headers_map": None}
        client_cls = functools.partial(
                CustomTestClient, connect_response_fields=connect_response_fields)
    else:
        client_cls = functools.partial(
                CustomTestClient, connect_response_fields=DEFAULT)

    protocol = AgentProtocol.connect(
            'app_name',
            LINKED_APPLICATIONS,
            ENVIRONMENT,
            global_settings(),
            client_cls=client_cls)

    protocol.send("shutdown")

    headers = protocol.client.headers[-1]
    assert headers == {
        "Content-Type": "application/json",
    }


def test_blob():
    request_headers_map = {u'X-Foo': u'Bar'}
    connect_response_fields = {u"request_headers_map": request_headers_map}

    client_cls = functools.partial(
            CustomTestClient,
            connect_response_fields=connect_response_fields)

    protocol = AgentProtocol.connect(
            'app_name',
            LINKED_APPLICATIONS,
            ENVIRONMENT,
            global_settings(),
            client_cls=client_cls)

    protocol.send("shutdown")

    headers = protocol.client.headers[-1]
    assert headers == {
        "Content-Type": "application/json",
        "X-Foo": "Bar",
    }


@override_generic_settings(global_settings(), {
    'developer_mode': True,
})
def test_server_side_config_precedence():
    connect_response_fields = {
        u'agent_config': {u'span_events.enabled': True},
        u'span_events.enabled': False,
    }
    client_cls = functools.partial(
            CustomTestClient,
            connect_response_fields=connect_response_fields)

    protocol = AgentProtocol.connect(
            'app_name',
            LINKED_APPLICATIONS,
            ENVIRONMENT,
            global_settings(),
            client_cls=client_cls)

    assert protocol.configuration.span_events.enabled is False


@override_generic_settings(global_settings(), {
    'developer_mode': True,
})
@pytest.mark.parametrize("connect_response_fields",
(
    {}, 
    {"span_event_harvest_config": {"report_period_ms": 60000, "harvest_limit": 123}}, 
    {"span_event_harvest_config": {}})
)
def test_span_event_harvest_config(connect_response_fields):
    client_cls = functools.partial(
            CustomTestClient,
            connect_response_fields=connect_response_fields)

    protocol = AgentProtocol.connect(
            'app_name',
            LINKED_APPLICATIONS,
            ENVIRONMENT,
            global_settings(),
            client_cls=client_cls)

    if connect_response_fields and connect_response_fields["span_event_harvest_config"]:
        expected = 123
    else:
        from newrelic.core.config import SPAN_EVENT_RESERVOIR_SIZE
        expected = SPAN_EVENT_RESERVOIR_SIZE
    assert protocol.configuration.event_harvest_config.harvest_limits.span_event_data == expected



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from newrelic.common.object_wrapper import transient_function_wrapper


def validate_database_duration():
    @transient_function_wrapper('newrelic.core.stats_engine',
            'StatsEngine.record_transaction')
    def _validate_database_duration(wrapped, instance, args, kwargs):
        try:
            result = wrapped(*args, **kwargs)
        except:
            raise
        else:

            metrics = instance.stats_table
            transaction_events = instance.transaction_events

            assert transaction_events.num_seen == 1

            event = next(iter(transaction_events))
            intrinsics = event[0]

            # As long as we are sending 'Database' metrics, then
            # 'databaseDuration' and 'databaseCallCount' will be
            # the sum both 'Database' and 'Datastore' values.

            try:
                database_all = metrics[('Database/all', '')]
            except KeyError:
                database_all_duration = 0.0
                database_all_call_count = 0
            else:
                database_all_duration = database_all.total_call_time
                database_all_call_count = database_all.call_count

            try:
                datastore_all = metrics[('Datastore/all', '')]
            except KeyError:
                datastore_all_duration = 0.0
                datastore_all_call_count = 0
            else:
                datastore_all_duration = datastore_all.total_call_time
                datastore_all_call_count = datastore_all.call_count

            assert 'databaseDuration' in intrinsics
            assert 'databaseCallCount' in intrinsics

            assert intrinsics['databaseDuration'] == (database_all_duration +
                    datastore_all_duration)
            assert intrinsics['databaseCallCount'] == (
                    database_all_call_count + datastore_all_call_count)

        return result

    return _validate_database_duration



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import uuid

import pytest

from google.cloud.firestore import Client
from google.cloud.firestore import Client, AsyncClient

from testing_support.db_settings import firestore_settings
from testing_support.fixture.event_loop import event_loop as loop  # noqa: F401; pylint: disable=W0611
from testing_support.fixtures import (  # noqa: F401; pylint: disable=W0611
    collector_agent_registration_fixture,
    collector_available_fixture,
)

from newrelic.api.datastore_trace import DatastoreTrace
from newrelic.api.time_trace import current_trace
from newrelic.common.system_info import LOCALHOST_EQUIVALENTS, gethostname

DB_SETTINGS = firestore_settings()[0]
FIRESTORE_HOST = DB_SETTINGS["host"]
FIRESTORE_PORT = DB_SETTINGS["port"]

_default_settings = {
    "transaction_tracer.explain_threshold": 0.0,
    "transaction_tracer.transaction_threshold": 0.0,
    "transaction_tracer.stack_trace_threshold": 0.0,
    "debug.log_data_collector_payloads": True,
    "debug.record_transaction_failure": True,
    "debug.log_explain_plan_queries": True,
}

collector_agent_registration = collector_agent_registration_fixture(
    app_name="Python Agent Test (datastore_firestore)",
    default_settings=_default_settings,
    linked_applications=["Python Agent Test (datastore)"],
)


@pytest.fixture()
def instance_info():
    host = gethostname() if FIRESTORE_HOST in LOCALHOST_EQUIVALENTS else FIRESTORE_HOST
    return {"host": host, "port_path_or_id": str(FIRESTORE_PORT), "db.instance": "projects/google-cloud-firestore-emulator/databases/(default)"}


@pytest.fixture(scope="session")
def client():
    os.environ["FIRESTORE_EMULATOR_HOST"] = "%s:%d" % (FIRESTORE_HOST, FIRESTORE_PORT)
    client = Client()
    # Ensure connection is available
    client.collection("healthcheck").document("healthcheck").set(
        {}, retry=None, timeout=5
    )
    return client


@pytest.fixture(scope="function")
def collection(client):
    collection_ = client.collection("firestore_collection_" + str(uuid.uuid4()))
    yield collection_
    client.recursive_delete(collection_)


@pytest.fixture(scope="session")
def async_client(loop):
    os.environ["FIRESTORE_EMULATOR_HOST"] = "%s:%d" % (FIRESTORE_HOST, FIRESTORE_PORT)
    client = AsyncClient()
    loop.run_until_complete(client.collection("healthcheck").document("healthcheck").set({}, retry=None, timeout=5))  # Ensure connection is available
    return client


@pytest.fixture(scope="function")
def async_collection(async_client, collection):
    # Use the same collection name as the collection fixture
    yield async_client.collection(collection.id)


@pytest.fixture(scope="session")
def assert_trace_for_generator():
    def _assert_trace_for_generator(generator_func, *args, **kwargs):
        txn = current_trace()
        assert not isinstance(txn, DatastoreTrace)

        # Check for generator trace on collections
        _trace_check = []
        for _ in generator_func(*args, **kwargs):
            _trace_check.append(isinstance(current_trace(), DatastoreTrace))
        assert _trace_check and all(_trace_check)  # All checks are True, and at least 1 is present.
        assert current_trace() is txn  # Generator trace has exited.

    return _assert_trace_for_generator


@pytest.fixture(scope="session")
def assert_trace_for_async_generator(loop):
    def _assert_trace_for_async_generator(generator_func, *args, **kwargs):
        _trace_check = []
        txn = current_trace()
        assert not isinstance(txn, DatastoreTrace)
        
        async def coro():
            # Check for generator trace on collections
            async for _ in generator_func(*args, **kwargs):
                _trace_check.append(isinstance(current_trace(), DatastoreTrace))

        loop.run_until_complete(coro())

        assert _trace_check and all(_trace_check)  # All checks are True, and at least 1 is present.
        assert current_trace() is txn  # Generator trace has exited.

    return _assert_trace_for_async_generator



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pika
from conftest import BODY, CORRELATION_ID, EXCHANGE, HEADERS, QUEUE, REPLY_TO
from testing_support.db_settings import rabbitmq_settings
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)
from testing_support.validators.validate_tt_collector_json import (
    validate_tt_collector_json,
)

from newrelic.api.background_task import background_task

DB_SETTINGS = rabbitmq_settings()[0]

_message_broker_tt_params = {
    "queue_name": QUEUE,
    "routing_key": QUEUE,
    "correlation_id": CORRELATION_ID,
    "reply_to": REPLY_TO,
    "headers": HEADERS.copy(),
}

_test_blocking_connection_consume_metrics = [
    ("MessageBroker/RabbitMQ/Exchange/Produce/Named/%s" % EXCHANGE, None),
    ("MessageBroker/RabbitMQ/Exchange/Consume/Named/%s" % EXCHANGE, None),
    ("MessageBroker/RabbitMQ/Exchange/Consume/Named/Unknown", None),
]


@validate_transaction_metrics(
    ("test_pika_blocking_connection_consume_generator:" "test_blocking_connection_consume_break"),
    scoped_metrics=_test_blocking_connection_consume_metrics,
    rollup_metrics=_test_blocking_connection_consume_metrics,
    background_task=True,
)
@validate_tt_collector_json(message_broker_params=_message_broker_tt_params)
@background_task()
def test_blocking_connection_consume_break(producer):
    with pika.BlockingConnection(pika.ConnectionParameters(DB_SETTINGS["host"])) as connection:
        channel = connection.channel()
        for method_frame, properties, body in channel.consume(QUEUE):
            assert hasattr(method_frame, "_nr_start_time")
            assert body == BODY
            break


@validate_transaction_metrics(
    ("test_pika_blocking_connection_consume_generator:" "test_blocking_connection_consume_connection_close"),
    scoped_metrics=_test_blocking_connection_consume_metrics,
    rollup_metrics=_test_blocking_connection_consume_metrics,
    background_task=True,
)
@validate_tt_collector_json(message_broker_params=_message_broker_tt_params)
@background_task()
def test_blocking_connection_consume_connection_close(producer):
    connection = pika.BlockingConnection(pika.ConnectionParameters(DB_SETTINGS["host"]))
    channel = connection.channel()

    try:
        for method_frame, properties, body in channel.consume(QUEUE):
            assert hasattr(method_frame, "_nr_start_time")
            assert body == BODY
            channel.close()
            connection.close()
    except pika.exceptions.ConnectionClosed:
        pass
    except:
        connection.close()


@validate_transaction_metrics(
    ("test_pika_blocking_connection_consume_generator:" "test_blocking_connection_consume_timeout"),
    scoped_metrics=_test_blocking_connection_consume_metrics,
    rollup_metrics=_test_blocking_connection_consume_metrics,
    background_task=True,
)
@validate_tt_collector_json(message_broker_params=_message_broker_tt_params)
@background_task()
def test_blocking_connection_consume_timeout(producer):
    with pika.BlockingConnection(pika.ConnectionParameters(DB_SETTINGS["host"])) as connection:
        channel = connection.channel()

        for result in channel.consume(QUEUE, inactivity_timeout=0.01):
            # result is None if there is a timeout
            if result and any(result):
                method_frame, properties, body = result
                channel.basic_ack(method_frame.delivery_tag)
                assert hasattr(method_frame, "_nr_start_time")
                assert body == BODY
            else:
                # timeout hit!
                break


@validate_transaction_metrics(
    ("test_pika_blocking_connection_consume_generator:" "test_blocking_connection_consume_exception_in_for_loop"),
    scoped_metrics=_test_blocking_connection_consume_metrics,
    rollup_metrics=_test_blocking_connection_consume_metrics,
    background_task=True,
)
@validate_tt_collector_json(message_broker_params=_message_broker_tt_params)
@background_task()
def test_blocking_connection_consume_exception_in_for_loop(producer):
    with pika.BlockingConnection(pika.ConnectionParameters(DB_SETTINGS["host"])) as connection:
        channel = connection.channel()

        try:
            # We should still create the metric in this case even if there is
            # an exception
            for result in channel.consume(QUEUE):
                1 / 0
        except ZeroDivisionError:
            # Expected error
            pass
        except Exception as e:
            assert False, "Wrong exception was raised: %s" % e
        else:
            assert False, "No exception was raised!"


_test_blocking_connection_consume_empty_metrics = [
    ("MessageBroker/RabbitMQ/Exchange/Produce/Named/%s" % EXCHANGE, None),
    ("MessageBroker/RabbitMQ/Exchange/Consume/Named/%s" % EXCHANGE, None),
    ("MessageBroker/RabbitMQ/Exchange/Consume/Named/Unknown", None),
]


@validate_transaction_metrics(
    ("test_pika_blocking_connection_consume_generator:" "test_blocking_connection_consume_exception_in_generator"),
    scoped_metrics=_test_blocking_connection_consume_empty_metrics,
    rollup_metrics=_test_blocking_connection_consume_empty_metrics,
    background_task=True,
)
@validate_tt_collector_json(message_broker_params=_message_broker_tt_params)
@background_task()
def test_blocking_connection_consume_exception_in_generator():
    with pika.BlockingConnection(pika.ConnectionParameters(DB_SETTINGS["host"])) as connection:
        channel = connection.channel()

        try:
            # Since the pytest fixture is not used, the QUEUE will not exist
            for result in channel.consume(QUEUE):
                pass
        except pika.exceptions.ChannelClosed:
            # Expected error
            pass
        except Exception as e:
            assert False, "Wrong exception was raised: %s" % e
        else:
            assert False, "No exception was raised!"


_test_blocking_connection_consume_many_metrics = [
    ("MessageBroker/RabbitMQ/Exchange/Produce/Named/%s" % EXCHANGE, None),
    ("MessageBroker/RabbitMQ/Exchange/Consume/Named/%s" % EXCHANGE, None),
    ("MessageBroker/RabbitMQ/Exchange/Consume/Named/Unknown", None),
]


@validate_transaction_metrics(
    ("test_pika_blocking_connection_consume_generator:" "test_blocking_connection_consume_many"),
    scoped_metrics=_test_blocking_connection_consume_many_metrics,
    rollup_metrics=_test_blocking_connection_consume_many_metrics,
    background_task=True,
)
@validate_tt_collector_json(message_broker_params=_message_broker_tt_params)
@background_task()
def test_blocking_connection_consume_many(produce_five):
    with pika.BlockingConnection(pika.ConnectionParameters(DB_SETTINGS["host"])) as connection:
        channel = connection.channel()

        consumed = 0
        for result in channel.consume(QUEUE, inactivity_timeout=0.01):
            if result and any(result):
                consumed += 1
            else:
                assert consumed == 5
                break


@validate_transaction_metrics(
    ("test_pika_blocking_connection_consume_generator:" "test_blocking_connection_consume_using_methods"),
    scoped_metrics=_test_blocking_connection_consume_metrics,
    rollup_metrics=_test_blocking_connection_consume_metrics,
    background_task=True,
)
@validate_tt_collector_json(message_broker_params=_message_broker_tt_params)
@background_task()
def test_blocking_connection_consume_using_methods(producer):
    with pika.BlockingConnection(pika.ConnectionParameters(DB_SETTINGS["host"])) as connection:
        channel = connection.channel()

        consumer = channel.consume(QUEUE, inactivity_timeout=0.01)

        method, properties, body = next(consumer)
        assert hasattr(method, "_nr_start_time")
        assert body == BODY

        result = next(consumer)
        assert result is None or not any(result)

        try:
            consumer.throw(ZeroDivisionError)
        except ZeroDivisionError:
            # This is expected
            pass
        else:
            # this is not
            assert False, "No exception was raised!"

        result = consumer.close()
        assert result is None


@validate_transaction_metrics(
    "Named/%s" % EXCHANGE,
    scoped_metrics=_test_blocking_connection_consume_metrics,
    rollup_metrics=_test_blocking_connection_consume_metrics,
    background_task=True,
    group="Message/RabbitMQ/Exchange",
)
@validate_tt_collector_json(message_broker_params=_message_broker_tt_params)
def test_blocking_connection_consume_outside_txn(producer):
    with pika.BlockingConnection(pika.ConnectionParameters(DB_SETTINGS["host"])) as connection:
        channel = connection.channel()
        consumer = channel.consume(QUEUE)

        try:
            for method_frame, properties, body in consumer:
                assert hasattr(method_frame, "_nr_start_time")
                assert body == BODY
                break
        finally:
            # Required for PyPy compatibility, see http://pypy.org/compat.html
            consumer.close()


def test_blocking_connection_consume_many_outside_txn(produce_five):
    @validate_transaction_metrics(
        "Named/%s" % EXCHANGE,
        scoped_metrics=_test_blocking_connection_consume_metrics,
        rollup_metrics=_test_blocking_connection_consume_metrics,
        background_task=True,
        group="Message/RabbitMQ/Exchange",
    )
    @validate_tt_collector_json(message_broker_params=_message_broker_tt_params)
    def consume_it(consumer, up_next=None):
        if up_next is None:
            method_frame, properties, body = next(consumer)
        else:
            method_frame, properties, body = up_next
        assert hasattr(method_frame, "_nr_start_time")
        assert body == BODY
        return next(consumer)

    with pika.BlockingConnection(pika.ConnectionParameters(DB_SETTINGS["host"])) as connection:
        channel = connection.channel()
        consumer = channel.consume(QUEUE)

        up_next = None
        for i in range(6):
            try:
                up_next = consume_it(consumer, up_next=up_next)
            except StopIteration:
                pass
            finally:
                consumer.close()


@validate_transaction_metrics(
    "Named/%s" % EXCHANGE,
    scoped_metrics=_test_blocking_connection_consume_metrics,
    rollup_metrics=_test_blocking_connection_consume_metrics,
    background_task=True,
    group="Message/RabbitMQ/Exchange",
)
@validate_tt_collector_json(message_broker_params=_message_broker_tt_params)
def test_blocking_connection_consume_using_methods_outside_txn(producer):
    with pika.BlockingConnection(pika.ConnectionParameters(DB_SETTINGS["host"])) as connection:
        channel = connection.channel()

        consumer = channel.consume(QUEUE, inactivity_timeout=0.01)

        method, properties, body = next(consumer)
        assert hasattr(method, "_nr_start_time")
        assert body == BODY

        result = next(consumer)
        assert result is None or not any(result)

        try:
            consumer.throw(ZeroDivisionError)
        except ZeroDivisionError:
            # This is expected
            pass
        else:
            # this is not
            assert False, "No exception was raised!"

        result = consumer.close()
        assert result is None


@validate_transaction_metrics(
    ("test_pika_blocking_connection_consume_generator:" "test_blocking_connection_consume_exception_on_creation"),
    scoped_metrics=_test_blocking_connection_consume_empty_metrics,
    rollup_metrics=_test_blocking_connection_consume_empty_metrics,
    background_task=True,
)
@background_task()
def test_blocking_connection_consume_exception_on_creation():
    with pika.BlockingConnection(pika.ConnectionParameters(DB_SETTINGS["host"])) as connection:
        channel = connection.channel()

        try:
            channel.consume(kittens=True)
        except TypeError:
            # this is expected
            pass
        else:
            # this is not
            assert False, "TypeError was not raised"



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest
from framework_graphql.test_application import *

from newrelic.common.package_version_utils import get_package_version

ARIADNE_VERSION = get_package_version("ariadne")
ariadne_version_tuple = tuple(map(int, ARIADNE_VERSION.split(".")))


@pytest.fixture(
    scope="session", params=["sync-sync", "async-sync", "async-async", "wsgi-sync", "asgi-sync", "asgi-async"]
)
def target_application(request):
    from ._target_application import target_application

    target_application = target_application[request.param]

    param = request.param.split("-")
    is_background = param[0] not in {"wsgi", "asgi"}
    schema_type = param[1]
    extra_spans = 4 if param[0] == "wsgi" else 0

    assert ARIADNE_VERSION is not None
    return "Ariadne", ARIADNE_VERSION, target_application, is_background, schema_type, extra_spans



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import uuid

import pytest
from testing_support.db_settings import kafka_settings

from testing_support.fixtures import collector_agent_registration_fixture, collector_available_fixture  # noqa: F401; pylint: disable=W0611

from newrelic.api.transaction import current_transaction
from newrelic.common.object_wrapper import transient_function_wrapper

DB_SETTINGS = kafka_settings()[0]

BROKER = "%s:%s" % (DB_SETTINGS["host"], DB_SETTINGS["port"])


_default_settings = {
    "transaction_tracer.explain_threshold": 0.0,
    "transaction_tracer.transaction_threshold": 0.0,
    "transaction_tracer.stack_trace_threshold": 0.0,
    "debug.log_data_collector_payloads": True,
    "debug.record_transaction_failure": True,
}

collector_agent_registration = collector_agent_registration_fixture(
    app_name="Python Agent Test (messagebroker_confluentkafka)",
    default_settings=_default_settings,
    linked_applications=["Python Agent Test (messagebroker_confluentkafka)"],
)


@pytest.fixture(scope="session", params=["cimpl", "serializer_function", "serializer_object"])
def client_type(request):
    return request.param


@pytest.fixture()
def skip_if_not_serializing(client_type):
    if client_type == "cimpl":
        pytest.skip("Only serializing clients supported.")


@pytest.fixture(scope="function")
def producer(topic, client_type, json_serializer):
    from confluent_kafka import Producer, SerializingProducer

    if client_type == "cimpl":
        producer = Producer({"bootstrap.servers": BROKER})
    elif client_type == "serializer_function":
        producer = SerializingProducer(
            {
                "bootstrap.servers": BROKER,
                "value.serializer": lambda v, c: json.dumps(v).encode("utf-8"),
                "key.serializer": lambda v, c: json.dumps(v).encode("utf-8") if v is not None else None,
            }
        )
    elif client_type == "serializer_object":
        producer = SerializingProducer(
            {
                "bootstrap.servers": BROKER,
                "value.serializer": json_serializer,
                "key.serializer": json_serializer,
            }
        )

    yield producer

    if hasattr(producer, "purge"):
        producer.purge()


@pytest.fixture(scope="function")
def consumer(group_id, topic, producer, client_type, json_deserializer):
    from confluent_kafka import Consumer, DeserializingConsumer

    if client_type == "cimpl":
        consumer = Consumer(
            {
                "bootstrap.servers": BROKER,
                "auto.offset.reset": "earliest",
                "heartbeat.interval.ms": 1000,
                "group.id": group_id,
            }
        )
    elif client_type == "serializer_function":
        consumer = DeserializingConsumer(
            {
                "bootstrap.servers": BROKER,
                "auto.offset.reset": "earliest",
                "heartbeat.interval.ms": 1000,
                "group.id": group_id,
                "value.deserializer": lambda v, c: json.loads(v.decode("utf-8")),
                "key.deserializer": lambda v, c: json.loads(v.decode("utf-8")) if v is not None else None,
            }
        )
    elif client_type == "serializer_object":
        consumer = DeserializingConsumer(
            {
                "bootstrap.servers": BROKER,
                "auto.offset.reset": "earliest",
                "heartbeat.interval.ms": 1000,
                "group.id": group_id,
                "value.deserializer": json_deserializer,
                "key.deserializer": json_deserializer,
            }
        )

    consumer.subscribe([topic])

    yield consumer

    consumer.close()


@pytest.fixture(scope="session")
def serialize(client_type):
    if client_type == "cimpl":
        return lambda v: json.dumps(v).encode("utf-8")
    else:
        return lambda v: v


@pytest.fixture(scope="session")
def deserialize(client_type):
    if client_type == "cimpl":
        return lambda v: json.loads(v.decode("utf-8"))
    else:
        return lambda v: v


@pytest.fixture(scope="session")
def json_serializer():
    from confluent_kafka.serialization import Serializer

    class JSONSerializer(Serializer):
        def __call__(self, obj, ctx):
            return json.dumps(obj).encode("utf-8") if obj is not None else None

    return JSONSerializer()


@pytest.fixture(scope="session")
def json_deserializer():
    from confluent_kafka.serialization import Deserializer

    class JSONDeserializer(Deserializer):
        def __call__(self, obj, ctx):
            return json.loads(obj.decode("utf-8")) if obj is not None else None

    return JSONDeserializer()


@pytest.fixture(scope="function")
def topic():
    from confluent_kafka.admin import AdminClient, NewTopic

    topic = "test-topic-%s" % str(uuid.uuid4())

    admin = AdminClient({"bootstrap.servers": BROKER})
    new_topics = [NewTopic(topic, num_partitions=1, replication_factor=1)]
    topics = admin.create_topics(new_topics)
    for _, f in topics.items():
        f.result()  # Block until topic is created.

    yield topic

    admin.delete_topics(new_topics)


@pytest.fixture(scope="session")
def group_id():
    return str(uuid.uuid4())


@pytest.fixture()
def send_producer_message(topic, producer, serialize, client_type):
    callback_called = []

    def producer_callback(err, msg):
        callback_called.append(True)

    def _test():
        if client_type == "cimpl":
            producer.produce(topic, value=serialize({"foo": 1}), callback=producer_callback)
        else:
            producer.produce(topic, value=serialize({"foo": 1}), on_delivery=producer_callback)
        producer.flush()
        assert callback_called

    return _test


@pytest.fixture()
def get_consumer_record(topic, send_producer_message, consumer, deserialize):
    def _test():
        send_producer_message()

        record_count = 0

        timeout = 10
        attempts = 0
        record = None
        while not record and attempts < timeout:
            record = consumer.poll(0.5)
            if not record:
                attempts += 1
                continue
            assert not record.error()

            assert deserialize(record.value()) == {"foo": 1}
            record_count += 1
        consumer.poll(0.5)  # Exit the transaction.

        assert record_count == 1, "Incorrect count of records consumed: %d. Expected 1." % record_count

    return _test


def cache_kafka_producer_headers():
    import confluent_kafka.cimpl

    @transient_function_wrapper(confluent_kafka.cimpl, "Producer.produce.__wrapped__")
    # Place transient wrapper underneath instrumentation
    def _cache_kafka_producer_headers(wrapped, instance, args, kwargs):
        transaction = current_transaction()

        if transaction is None:
            return wrapped(*args, **kwargs)

        ret = wrapped(*args, **kwargs)
        headers = kwargs.get("headers", [])
        headers = dict(headers)
        transaction._test_request_headers = headers
        return ret

    return _cache_kafka_producer_headers


def cache_kafka_consumer_headers():
    import confluent_kafka.cimpl

    @transient_function_wrapper(confluent_kafka.cimpl, "Consumer.poll")
    # Place transient wrapper underneath instrumentation
    def _cache_kafka_consumer_headers(wrapped, instance, args, kwargs):
        record = wrapped(*args, **kwargs)
        transaction = current_transaction()

        if transaction is None:
            return record

        headers = dict(record.headers())
        transaction._test_request_headers = headers
        return record

    return _cache_kafka_consumer_headers


@pytest.fixture(autouse=True)
def assert_no_active_transaction():
    # Run before test
    assert not current_transaction(active_only=False), "Transaction exists before test run."

    yield  # Run test

    # Run after test
    assert not current_transaction(active_only=False), "Transaction was not properly exited."



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This is an implementation of the cross agent tests for cat map using a wsgi
application. Another implementation of these tests using a tornado application
can be found in test/framework_tornado_r3/test_cat_map.py
"""

import json
import os

import pytest
import webtest

try:
    from urllib2 import urlopen  # Py2.X
except ImportError:
    from urllib.request import urlopen  # Py3.X

from testing_support.fixtures import (
    make_cross_agent_headers,
    override_application_name,
    override_application_settings,
    validate_analytics_catmap_data,
)
from testing_support.mock_external_http_server import (
    MockExternalHTTPHResponseHeadersServer,
)
from testing_support.validators.validate_tt_parameters import validate_tt_parameters

from newrelic.api.external_trace import ExternalTrace
from newrelic.api.transaction import (
    current_transaction,
    get_browser_timing_header,
    set_background_task,
    set_transaction_name,
)
from newrelic.api.wsgi_application import wsgi_application
from newrelic.common.encoding_utils import json_encode, obfuscate
from newrelic.packages import six

ENCODING_KEY = "1234567890123456789012345678901234567890"
CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
JSON_DIR = os.path.normpath(os.path.join(CURRENT_DIR, "fixtures"))
OUTBOUD_REQUESTS = {}

_parameters_list = [
    "name",
    "appName",
    "transactionName",
    "transactionGuid",
    "inboundPayload",
    "outboundRequests",
    "expectedIntrinsicFields",
    "nonExpectedIntrinsicFields",
]


@pytest.fixture(scope="module")
def server():
    with MockExternalHTTPHResponseHeadersServer() as _server:
        yield _server


def load_tests():
    result = []
    path = os.path.join(JSON_DIR, "cat_map.json")
    with open(path, "r") as fh:
        tests = json.load(fh)

    for test in tests:
        values = tuple([test.get(param, None) for param in _parameters_list])
        result.append(values)

    return result


_parameters = ",".join(_parameters_list)


@wsgi_application()
def target_wsgi_application(environ, start_response):
    status = "200 OK"

    txn_name = environ.get("txn")
    if six.PY2:
        txn_name = txn_name.decode("UTF-8")
    txn_name = txn_name.split("/", 3)

    guid = environ.get("guid")
    old_cat = environ.get("old_cat") == "True"
    txn = current_transaction()

    txn.guid = guid
    for req in OUTBOUD_REQUESTS:
        # Change the transaction name before making an outbound call.
        outgoing_name = req["outboundTxnName"].split("/", 3)
        if outgoing_name[0] != "WebTransaction":
            set_background_task(True)

        set_transaction_name(outgoing_name[2], group=outgoing_name[1])

        expected_outbound_header = obfuscate(json_encode(req["expectedOutboundPayload"]), ENCODING_KEY)
        generated_outbound_header = dict(ExternalTrace.generate_request_headers(txn))

        # A 500 error is returned because 'assert' statements in the wsgi app
        # are ignored.

        if old_cat:
            if expected_outbound_header != generated_outbound_header["X-NewRelic-Transaction"]:
                status = "500 Outbound Headers Check Failed."
        else:
            if "X-NewRelic-Transaction" in generated_outbound_header:
                status = "500 Outbound Headers Check Failed."
        r = urlopen(environ["server_url"])  # nosec B310
        r.read(10)

    # Set the final transaction name.

    if txn_name[0] != "WebTransaction":
        set_background_task(True)
    set_transaction_name(txn_name[2], group=txn_name[1])

    text = "<html><head>%s</head><body><p>RESPONSE</p></body></html>"

    output = (text % get_browser_timing_header()).encode("UTF-8")

    response_headers = [("Content-type", "text/html; charset=utf-8"), ("Content-Length", str(len(output)))]
    start_response(status, response_headers)

    return [output]


target_application = webtest.TestApp(target_wsgi_application)


@pytest.mark.parametrize(_parameters, load_tests())
@pytest.mark.parametrize("old_cat", (True, False))
def test_cat_map(
    name,
    appName,
    transactionName,
    transactionGuid,
    inboundPayload,
    outboundRequests,
    expectedIntrinsicFields,
    nonExpectedIntrinsicFields,
    old_cat,
    server,
):
    global OUTBOUD_REQUESTS
    OUTBOUD_REQUESTS = outboundRequests or {}

    _custom_settings = {
        "cross_process_id": "1#1",
        "encoding_key": ENCODING_KEY,
        "trusted_account_ids": [1],
        "cross_application_tracer.enabled": True,
        "distributed_tracing.enabled": not old_cat,
        "transaction_tracer.transaction_threshold": 0.0,
    }

    if expectedIntrinsicFields and old_cat:
        _external_node_params = {
            "path_hash": expectedIntrinsicFields["nr.pathHash"],
            "trip_id": expectedIntrinsicFields["nr.tripId"],
        }
    else:
        _external_node_params = []

    if not old_cat:
        # since no better cat headers will be generated, no intrinsics should
        # be added
        expectedIntrinsicFields = {}

    @validate_tt_parameters(required_params=_external_node_params)
    @validate_analytics_catmap_data(
        transactionName, expected_attributes=expectedIntrinsicFields, non_expected_attributes=nonExpectedIntrinsicFields
    )
    @override_application_settings(_custom_settings)
    @override_application_name(appName)
    def run_cat_test():
        if six.PY2:
            txn_name = transactionName.encode("UTF-8")
            guid = transactionGuid.encode("UTF-8")
        else:
            txn_name = transactionName
            guid = transactionGuid

        # Only generate old cat style headers. This will test to make sure we
        # are properly ignoring these headers when the agent is using better
        # cat.

        headers = make_cross_agent_headers(inboundPayload, ENCODING_KEY, "1#1")
        response = target_application.get(
            "/",
            headers=headers,
            extra_environ={
                "txn": txn_name,
                "guid": guid,
                "old_cat": str(old_cat),
                "server_url": "http://localhost:%d" % server.port,
            },
        )

        # Validation of analytic data happens in the decorator.

        assert response.status == "200 OK"

        content = response.html.html.body.p.string

        # Validate actual body content as sanity check.

        assert content == "RESPONSE"

    run_cat_test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Union

import strawberry.mutation
import strawberry.type
from strawberry import Schema, field
from strawberry.asgi import GraphQL
from strawberry.schema.config import StrawberryConfig
from strawberry.types.types import Optional
from testing_support.asgi_testing import AsgiTest


@strawberry.type
class Author:
    first_name: str
    last_name: str


@strawberry.type
class Book:
    id: int
    name: str
    isbn: str
    author: Author
    branch: str


@strawberry.type
class Magazine:
    id: int
    name: str
    issue: int
    branch: str


@strawberry.type
class Library:
    id: int
    branch: str
    magazine: List[Magazine]
    book: List[Book]


Item = Union[Book, Magazine]
Storage = List[str]


authors = [
    Author(
        first_name="New",
        last_name="Relic",
    ),
    Author(
        first_name="Bob",
        last_name="Smith",
    ),
    Author(
        first_name="Leslie",
        last_name="Jones",
    ),
]

books = [
    Book(
        id=1,
        name="Python Agent: The Book",
        isbn="a-fake-isbn",
        author=authors[0],
        branch="riverside",
    ),
    Book(
        id=2,
        name="Ollies for O11y: A Sk8er's Guide to Observability",
        isbn="a-second-fake-isbn",
        author=authors[1],
        branch="downtown",
    ),
    Book(
        id=3,
        name="[Redacted]",
        isbn="a-third-fake-isbn",
        author=authors[2],
        branch="riverside",
    ),
]

magazines = [
    Magazine(id=1, name="Reli Updates Weekly", issue=1, branch="riverside"),
    Magazine(id=2, name="Reli: The Forgotten Years", issue=2, branch="downtown"),
    Magazine(id=3, name="Node Weekly", issue=1, branch="riverside"),
]


libraries = ["riverside", "downtown"]
libraries = [
    Library(
        id=i + 1,
        branch=branch,
        magazine=[m for m in magazines if m.branch == branch],
        book=[b for b in books if b.branch == branch],
    )
    for i, branch in enumerate(libraries)
]

storage = []


def resolve_hello():
    return "Hello!"


def resolve_echo(echo: str):
    return echo


def resolve_library(index: int):
    return libraries[index]


def resolve_storage_add(string: str):
    storage.append(string)
    return string


def resolve_storage():
    return [storage.pop()]


def resolve_error():
    raise RuntimeError("Runtime Error!")


def resolve_search(contains: str):
    search_books = [b for b in books if contains in b.name]
    search_magazines = [m for m in magazines if contains in m.name]
    return search_books + search_magazines


@strawberry.type
class Query:
    library: Library = field(resolver=resolve_library)
    hello: str = field(resolver=resolve_hello)
    search: List[Item] = field(resolver=resolve_search)
    echo: str = field(resolver=resolve_echo)
    storage: Storage = field(resolver=resolve_storage)
    error: Optional[str] = field(resolver=resolve_error)
    error_non_null: str = field(resolver=resolve_error)


@strawberry.type
class Mutation:
    storage_add: str = strawberry.mutation(resolver=resolve_storage_add)


target_schema = Schema(query=Query, mutation=Mutation, config=StrawberryConfig(auto_camel_case=False))
target_asgi_application = AsgiTest(GraphQL(target_schema))



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import webtest
import pytest

from newrelic.api.transaction import current_transaction
from newrelic.api.external_trace import ExternalTrace
from newrelic.api.wsgi_application import wsgi_application
from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_span_events import (
    validate_span_events)
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_transaction_event_attributes import validate_transaction_event_attributes

@wsgi_application()
def target_wsgi_application(environ, start_response):
    start_response('200 OK', [('Content-Type', 'application/json')])
    txn = current_transaction()
    if txn._sampled is None:
        txn._sampled = True
        txn._priority = 1.2
    headers = ExternalTrace.generate_request_headers(txn)
    return [json.dumps(headers).encode('utf-8')]


test_application = webtest.TestApp(target_wsgi_application)


_override_settings = {
    'trusted_account_key': '1',
    'account_id': '1',
    'primary_application_id': '2',
    'distributed_tracing.enabled': True,
}


INBOUND_TRACEPARENT_ZERO_PARENT_ID = \
        '00-0af7651916cd43dd8448eb211c80319c-0000000000000000-01'
INBOUND_TRACEPARENT_ZERO_TRACE_ID = \
        '00-00000000000000000000000000000000-00f067aa0ba902b7-01'
INBOUND_TRACEPARENT = \
        '00-0af7651916cd43dd8448eb211c80319c-00f067aa0ba902b7-01'
INBOUND_TRACEPARENT_NEW_VERSION_EXTRA_FIELDS = \
        '01-0af7651916cd43dd8448eb211c80319c-00f067aa0ba902b7-01-extra-field'
INBOUND_TRACEPARENT_VERSION_ZERO_EXTRA_FIELDS = \
        '00-0af7651916cd43dd8448eb211c80319c-00f067aa0ba902b7-01-extra-field'
INBOUND_TRACESTATE = \
        'rojo=f06a0ba902b7,congo=t61rcWkgMzE'
LONG_TRACESTATE = \
        ','.join(["{}@rojo=f06a0ba902b7".format(x) for x in range(32)])
INBOUND_UNTRUSTED_NR_TRACESTATE = \
        ('2@nr=0-0-1345936-55632452-27jjj2d8890283b4-b28ce285632jjhl9-'
        '1-1.1273-1569367663277')
INBOUND_NR_TRACESTATE = \
        ('1@nr=0-0-1349956-41346604-27ddd2d8890283b4-b28be285632bbc0a-'
        '1-1.4-1569367663277')
INBOUND_NR_TRACESTATE_UNSAMPLED = \
        ('1@nr=0-0-1349956-41346604-27ddd2d8890283b4-b28be285632bbc0a-'
        '0-0.4-1569367663277')
INBOUND_NR_TRACESTATE_INVALID_TIMESTAMP = \
        ('1@nr=0-0-1349956-41346604-27ddd2d8890283b4-b28be285632bbc0a-'
        '0-0.4-timestamp')
INBOUND_NR_TRACESTATE_INVALID_PARENT_TYPE = \
        ('1@nr=0-parentType-1349956-41346604-'
         '27ddd2d8890283b4-b28be285632bbc0a-1-1.4-1569367663277')

INBOUND_TRACEPARENT_VERSION_FF = \
        'ff-0af7651916cd43dd8448eb211c80319c-00f067aa0ba902b7-01'
INBOUND_TRACEPARENT_VERSION_TOO_LONG = \
        '000-0af7651916cd43dd8448eb211c80319c-00f067aa0ba902b7-01'
INBOUND_TRACEPARENT_INVALID_TRACE_ID = \
        '00-0aF7651916cd43dd8448eb211c80319c-00f067aa0ba902b7-01'
INBOUND_TRACEPARENT_INVALID_PARENT_ID = \
        '00-0af7651916cd43dd8448eb211c80319c-00f067aa0Ba902b7-01'
INBOUND_TRACEPARENT_INVALID_FLAGS = \
        '00-0af7651916cd43dd8448eb211c80319c-00f067aa0ba902b7-x1'

_metrics = [("Supportability/TraceContext/Create/Success", 1)]


@pytest.mark.parametrize('inbound_nr_tracestate', (True, False))
@override_application_settings(_override_settings)
def test_tracestate_generation(inbound_nr_tracestate):

    headers = {
        'traceparent': INBOUND_TRACEPARENT,
        'tracestate': (
                INBOUND_NR_TRACESTATE_UNSAMPLED if inbound_nr_tracestate
                else INBOUND_TRACESTATE),
    }

    @validate_transaction_metrics("",
            group="Uri",
            rollup_metrics=_metrics)
    def _test():
        return test_application.get('/', headers=headers)

    response = _test()
    for header_name, header_value in response.json:
        if header_name == 'tracestate':
            break
    else:
        assert False, 'tracestate header not propagated'

    header_value = header_value.split(',', 1)[0]
    key, value = header_value.split('=', 2)
    assert key == '1@nr'

    fields = value.split('-')
    assert len(fields) == 9

    assert str(int(fields[8])) == fields[8]
    deterministic_fields = fields[:4] + fields[6:8]
    assert deterministic_fields == [
        '0',
        '0',
        '1',
        '2',
        '0' if inbound_nr_tracestate else '1',
        '0.4' if inbound_nr_tracestate else '1.2',
    ]


    assert len(fields[4]) == 16
    assert len(fields[5]) == 16


@pytest.mark.parametrize('inbound_tracestate,expected', (
    ('', None),
    (INBOUND_NR_TRACESTATE + "," + INBOUND_TRACESTATE, INBOUND_TRACESTATE),
    (INBOUND_TRACESTATE, INBOUND_TRACESTATE),
    (LONG_TRACESTATE + ',' + INBOUND_NR_TRACESTATE,
            ','.join("{}@rojo=f06a0ba902b7".format(x) for x in range(31))),
), ids=(
    'empty_inbound_payload',
    'nr_payload',
    'no_nr_payload',
    'long_payload',
))
@override_application_settings(_override_settings)
def test_tracestate_propagation(inbound_tracestate, expected):
    headers = {
        'traceparent': INBOUND_TRACEPARENT,
        'tracestate': inbound_tracestate
    }
    response = test_application.get('/', headers=headers)
    for header_name, header_value in response.json:
        if header_name == 'tracestate':
            break
    else:
        assert False, 'tracestate header not propagated'

    assert not header_value.endswith(',')
    if inbound_tracestate:
        _, propagated_tracestate = header_value.split(',', 1)
        assert propagated_tracestate == expected


@pytest.mark.parametrize('inbound_traceparent,span_events_enabled', (
    (True, True),
    (True, False),
    (False, True),
))
def test_traceparent_generation(inbound_traceparent, span_events_enabled):
    settings = _override_settings.copy()
    settings['span_events.enabled'] = span_events_enabled

    headers = {}
    if inbound_traceparent:
        headers['traceparent'] = INBOUND_TRACEPARENT

    @override_application_settings(settings)
    def _test():
        return test_application.get('/', headers=headers)

    response = _test()
    for header_name, header_value in response.json:
        if header_name == 'traceparent':
            break
    else:
        assert False, 'traceparent header not present'

    assert len(header_value) == 55
    assert header_value.startswith('00-')
    fields = header_value.split('-')
    assert len(fields) == 4
    if inbound_traceparent:
        assert fields[1] == '0af7651916cd43dd8448eb211c80319c'
        assert fields[2] != '00f067aa0ba902b7'
    else:
        assert len(fields[1]) == 32
    assert fields[3] in ('00', '01')


@pytest.mark.parametrize('traceparent,intrinsics,metrics', (
    (INBOUND_TRACEPARENT, {
            "traceId": "0af7651916cd43dd8448eb211c80319c",
            "parentSpanId": "00f067aa0ba902b7",
            "parent.transportType": "HTTP"},
            [("Supportability/TraceContext/TraceParent/Accept/Success", 1)]),
    (INBOUND_TRACEPARENT_NEW_VERSION_EXTRA_FIELDS, {
            "traceId": "0af7651916cd43dd8448eb211c80319c",
            "parentSpanId": "00f067aa0ba902b7",
            "parent.transportType": "HTTP"},
            [("Supportability/TraceContext/TraceParent/Accept/Success", 1)]),
    (INBOUND_TRACEPARENT + ' ', {
            "traceId": "0af7651916cd43dd8448eb211c80319c",
            "parentSpanId": "00f067aa0ba902b7",
            "parent.transportType": "HTTP"},
            [("Supportability/TraceContext/TraceParent/Accept/Success", 1)]),

    ('INVALID', {},
        [("Supportability/TraceContext/TraceParent/Parse/Exception", 1)]),
    ('xx-0', {},
        [("Supportability/TraceContext/TraceParent/Parse/Exception", 1)]),
    (INBOUND_TRACEPARENT_VERSION_FF, {},
        [("Supportability/TraceContext/TraceParent/Parse/Exception", 1)]),
    (INBOUND_TRACEPARENT_VERSION_ZERO_EXTRA_FIELDS, {},
        [("Supportability/TraceContext/TraceParent/Parse/Exception", 1)]),
    (INBOUND_TRACEPARENT[:-1], {},
        [("Supportability/TraceContext/TraceParent/Parse/Exception", 1)]),
    ('00-xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx', {},
        [("Supportability/TraceContext/TraceParent/Parse/Exception", 1)]),
    (INBOUND_TRACEPARENT_INVALID_TRACE_ID, {},
        [("Supportability/TraceContext/TraceParent/Parse/Exception", 1)]),
    (INBOUND_TRACEPARENT_INVALID_PARENT_ID, {},
        [("Supportability/TraceContext/TraceParent/Parse/Exception", 1)]),
    (INBOUND_TRACEPARENT_INVALID_FLAGS, {},
        [("Supportability/TraceContext/TraceParent/Parse/Exception", 1)]),
    (INBOUND_TRACEPARENT_ZERO_TRACE_ID, {},
        [("Supportability/TraceContext/TraceParent/Parse/Exception", 1)]),
    (INBOUND_TRACEPARENT_ZERO_PARENT_ID, {},
        [("Supportability/TraceContext/TraceParent/Parse/Exception", 1)]),
    (INBOUND_TRACEPARENT_VERSION_TOO_LONG, {},
        [("Supportability/TraceContext/TraceParent/Parse/Exception", 1)]),
))
@override_application_settings(_override_settings)
def test_inbound_traceparent_header(traceparent, intrinsics, metrics):
    exact = {'agent': {}, 'user': {}, 'intrinsic': intrinsics}

    @validate_transaction_event_attributes(exact_attrs=exact)
    @validate_transaction_metrics(
            "", group="Uri", rollup_metrics=metrics)
    def _test():
        test_application.get('/', headers={"traceparent": traceparent})

    _test()


@pytest.mark.parametrize('tracestate,intrinsics', (
    (INBOUND_TRACESTATE,
            {'tracingVendors': 'rojo,congo',
             'parentId': '00f067aa0ba902b7'}),
    (INBOUND_NR_TRACESTATE,
            {'trustedParentId': '27ddd2d8890283b4'}),
    ('garbage', {'parentId': '00f067aa0ba902b7'}),
    (INBOUND_TRACESTATE + ',' + INBOUND_NR_TRACESTATE,
            {'parentId': '00f067aa0ba902b7',
             'trustedParentId': '27ddd2d8890283b4',
             'tracingVendors': 'rojo,congo'}),
    (INBOUND_TRACESTATE + ',' + INBOUND_UNTRUSTED_NR_TRACESTATE,
            {'parentId': '00f067aa0ba902b7',
             'tracingVendors': 'rojo,congo,2@nr'}),
    ('rojo=12345,' + 'v' * 257 + '=x',
            {'tracingVendors': 'rojo'}),
    ('rojo=12345,k=' + 'v' * 257,
            {'tracingVendors': 'rojo'}),
))
@override_application_settings(_override_settings)
def test_inbound_tracestate_header(tracestate, intrinsics):

    nr_entry_count = 1 if '1@nr' not in tracestate else None

    metrics = [
        ('Supportability/TraceContext/TraceState/NoNrEntry', nr_entry_count),
    ]

    @validate_transaction_metrics(
            "", group="Uri", rollup_metrics=metrics)
    @validate_span_events(exact_intrinsics=intrinsics)
    def _test():
        test_application.get('/', headers={
            "traceparent": INBOUND_TRACEPARENT,
            "tracestate": tracestate,
        })

    _test()


@override_application_settings(_override_settings)
def test_w3c_tracestate_header():

    metrics = [
        ('Supportability/TraceContext/Accept/Success', 1),
    ]

    @validate_transaction_metrics(
        "", group="Uri", rollup_metrics=metrics)
    def _test():
        test_application.get('/', headers={
            "traceparent": INBOUND_TRACEPARENT,
            "tracestate": INBOUND_TRACESTATE,
        })

    _test()


@pytest.mark.parametrize('tracestate', (
    (INBOUND_NR_TRACESTATE_INVALID_TIMESTAMP),
    (INBOUND_NR_TRACESTATE_INVALID_PARENT_TYPE),
    ))
@override_application_settings(_override_settings)
def test_invalid_inbound_nr_tracestate_header(tracestate):

    metrics = [
        ('Supportability/TraceContext/TraceState/InvalidNrEntry', 1)
    ]
    @validate_transaction_metrics(
        "", group="Uri", rollup_metrics=metrics)
    def _test():
        test_application.get('/', headers={
            "traceparent": INBOUND_TRACEPARENT,
            "tracestate": tracestate,
        })

    _test()


@pytest.mark.parametrize('exclude_newrelic_header', (True, False))
def test_w3c_and_newrelic_headers_generated(exclude_newrelic_header):

    settings = _override_settings.copy()
    settings['distributed_tracing.exclude_newrelic_header'] = \
            exclude_newrelic_header

    @override_application_settings(settings)
    def _test():
        return test_application.get('/')

    response = _test()
    traceparent = None
    tracestate = None
    newrelic = None
    for header_name, header_value in response.json:
        if header_name == 'traceparent':
            traceparent = header_value
        elif header_name == 'tracestate':
            tracestate = header_value
        elif header_name == 'newrelic':
            newrelic = header_value

    assert traceparent
    assert tracestate

    if exclude_newrelic_header:
        assert newrelic is None
    else:
        assert newrelic



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import memcache

from newrelic.api.background_task import background_task

from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.db_settings import memcached_settings
from testing_support.util import instance_hostname

DB_MULTIPLE_SETTINGS = memcached_settings()

# Settings

_enable_instance_settings = {
    'datastore_tracer.instance_reporting.enabled': True,
}
_disable_instance_settings = {
    'datastore_tracer.instance_reporting.enabled': False,
}

# Metrics

_base_scoped_metrics = (
    ('Datastore/operation/Memcached/get_multi', 1),
    ('Datastore/operation/Memcached/set_multi', 1),
)

_base_rollup_metrics = (
    ('Datastore/all', 2),
    ('Datastore/allOther', 2),
    ('Datastore/Memcached/all', 2),
    ('Datastore/Memcached/allOther', 2),
    ('Datastore/operation/Memcached/set_multi', 1),
    ('Datastore/operation/Memcached/get_multi', 1),
)

_disable_scoped_metrics = list(_base_scoped_metrics)
_disable_rollup_metrics = list(_base_rollup_metrics)

_enable_scoped_metrics = list(_base_scoped_metrics)
_enable_rollup_metrics = list(_base_rollup_metrics)

if len(DB_MULTIPLE_SETTINGS) > 1:
    memcached_1 = DB_MULTIPLE_SETTINGS[0]
    memcached_2 = DB_MULTIPLE_SETTINGS[1]

    host_1 = instance_hostname(memcached_1['host'])
    port_1 = memcached_1['port']

    host_2 = instance_hostname(memcached_2['host'])
    port_2 = memcached_2['port']

    instance_metric_name_1 = 'Datastore/instance/Memcached/%s/%s' % (host_1,
            port_1)
    instance_metric_name_2 = 'Datastore/instance/Memcached/%s/%s' % (host_2,
            port_2)

    _enable_rollup_metrics.extend([
            (instance_metric_name_1, None),
            (instance_metric_name_2, None),
    ])

    _disable_rollup_metrics.extend([
            (instance_metric_name_1, None),
            (instance_metric_name_2, None),
    ])

def exercise_memcached(client, multi_dict):
    client.set_multi(multi_dict)
    client.get_multi(multi_dict.keys())

transaction_metric_prefix = 'test_multiple_dbs:test_multiple_datastores'

@pytest.mark.skipif(len(DB_MULTIPLE_SETTINGS) < 2,
        reason='Test environment not configured with multiple databases.')
@override_application_settings(_enable_instance_settings)
@validate_transaction_metrics(transaction_metric_prefix+'_enabled',
        scoped_metrics=_enable_scoped_metrics,
        rollup_metrics=_enable_rollup_metrics,
        background_task=True)
@background_task()
def test_multiple_datastores_enabled(memcached_multi):
    memcached1 = DB_MULTIPLE_SETTINGS[0]
    memcached2 = DB_MULTIPLE_SETTINGS[1]
    settings = [memcached1, memcached2]
    servers = ["%s:%s" % (x['host'], x['port']) for x in settings]

    client = memcache.Client(servers=servers)

    exercise_memcached(client, memcached_multi)

@pytest.mark.skipif(len(DB_MULTIPLE_SETTINGS) < 2,
        reason='Test environment not configured with multiple databases.')
@override_application_settings(_disable_instance_settings)
@validate_transaction_metrics(transaction_metric_prefix+'_disabled',
        scoped_metrics=_disable_scoped_metrics,
        rollup_metrics=_disable_rollup_metrics,
        background_task=True)
@background_task()
def test_multiple_datastores_disabled(memcached_multi):
    memcached1 = DB_MULTIPLE_SETTINGS[0]
    memcached2 = DB_MULTIPLE_SETTINGS[1]
    settings = [memcached1, memcached2]
    servers = ["%s:%s" % (x['host'], x['port']) for x in settings]

    client = memcache.Client(servers=servers)

    exercise_memcached(client, memcached_multi)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from testing_support.validators.validate_database_duration import (
    validate_database_duration,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)
from testing_support.validators.validate_tt_collector_json import (
    validate_tt_collector_json,
)

from newrelic.api.background_task import background_task


@pytest.fixture(autouse=True)
def sample_data(collection):
    for x in range(1, 4):
        collection.add({"x": x}, "doc%d" % x)


@pytest.fixture()
def exercise_async_transaction_commit(async_client, async_collection):
    async def _exercise_async_transaction_commit():
        from google.cloud.firestore import async_transactional

        @async_transactional
        async def _exercise(async_transaction):
            # get a DocumentReference
            with pytest.raises(
                TypeError
            ):  # get is currently broken. It attempts to await an async_generator instead of consuming it.
                [_ async for _ in async_transaction.get(async_collection.document("doc1"))]

            # get a Query
            with pytest.raises(
                TypeError
            ):  # get is currently broken. It attempts to await an async_generator instead of consuming it.
                async_query = async_collection.select("x").where(field_path="x", op_string=">", value=2)
                assert len([_ async for _ in async_transaction.get(async_query)]) == 1

            # get_all on a list of DocumentReferences
            with pytest.raises(
                TypeError
            ):  # get_all is currently broken. It attempts to await an async_generator instead of consuming it.
                all_docs = async_transaction.get_all([async_collection.document("doc%d" % x) for x in range(1, 4)])
                assert len([_ async for _ in all_docs]) == 3

            # set and delete methods
            async_transaction.set(async_collection.document("doc2"), {"x": 0})
            async_transaction.delete(async_collection.document("doc3"))

        await _exercise(async_client.transaction())
        assert len([_ async for _ in async_collection.list_documents()]) == 2

    return _exercise_async_transaction_commit


@pytest.fixture()
def exercise_async_transaction_rollback(async_client, async_collection):
    async def _exercise_async_transaction_rollback():
        from google.cloud.firestore import async_transactional

        @async_transactional
        async def _exercise(async_transaction):
            # set and delete methods
            async_transaction.set(async_collection.document("doc2"), {"x": 99})
            async_transaction.delete(async_collection.document("doc1"))
            raise RuntimeError()

        with pytest.raises(RuntimeError):
            await _exercise(async_client.transaction())
        assert len([_ async for _ in async_collection.list_documents()]) == 3

    return _exercise_async_transaction_rollback


def test_firestore_async_transaction_commit(loop, exercise_async_transaction_commit, async_collection, instance_info):
    _test_scoped_metrics = [
        ("Datastore/operation/Firestore/commit", 1),
        # ("Datastore/operation/Firestore/get_all", 2),
        # ("Datastore/statement/Firestore/%s/stream" % async_collection.id, 1),
        ("Datastore/statement/Firestore/%s/list_documents" % async_collection.id, 1),
    ]

    _test_rollup_metrics = [
        # ("Datastore/operation/Firestore/stream", 1),
        ("Datastore/operation/Firestore/list_documents", 1),
        ("Datastore/all", 2),  # Should be 5 if not for broken APIs
        ("Datastore/allOther", 2),
        ("Datastore/instance/Firestore/%s/%s" % (instance_info["host"], instance_info["port_path_or_id"]), 2),
    ]

    @validate_database_duration()
    @validate_transaction_metrics(
        "test_firestore_async_transaction",
        scoped_metrics=_test_scoped_metrics,
        rollup_metrics=_test_rollup_metrics,
        background_task=True,
    )
    @background_task(name="test_firestore_async_transaction")
    def _test():
        loop.run_until_complete(exercise_async_transaction_commit())

    _test()


def test_firestore_async_transaction_rollback(
    loop, exercise_async_transaction_rollback, async_collection, instance_info
):
    _test_scoped_metrics = [
        ("Datastore/operation/Firestore/rollback", 1),
        ("Datastore/statement/Firestore/%s/list_documents" % async_collection.id, 1),
    ]

    _test_rollup_metrics = [
        ("Datastore/operation/Firestore/list_documents", 1),
        ("Datastore/all", 2),
        ("Datastore/allOther", 2),
        ("Datastore/instance/Firestore/%s/%s" % (instance_info["host"], instance_info["port_path_or_id"]), 2),
    ]

    @validate_database_duration()
    @validate_transaction_metrics(
        "test_firestore_async_transaction",
        scoped_metrics=_test_scoped_metrics,
        rollup_metrics=_test_rollup_metrics,
        background_task=True,
    )
    @background_task(name="test_firestore_async_transaction")
    def _test():
        loop.run_until_complete(exercise_async_transaction_rollback())

    _test()


def test_firestore_async_transaction_commit_trace_node_datastore_params(
    loop, exercise_async_transaction_commit, instance_info
):
    @validate_tt_collector_json(datastore_params=instance_info)
    @background_task()
    def _test():
        loop.run_until_complete(exercise_async_transaction_commit())

    _test()


def test_firestore_async_transaction_rollback_trace_node_datastore_params(
    loop, exercise_async_transaction_rollback, instance_info
):
    @validate_tt_collector_json(datastore_params=instance_info)
    @background_task()
    def _test():
        loop.run_until_complete(exercise_async_transaction_rollback())

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from django.http import HttpResponseGone


class Custom410(Exception):
    pass


class ExceptionTo410Middleware(object):

    def __init__(self, get_response=None):
        self.get_response = get_response

    def __call__(self, request):
        return self.get_response(request)

    @staticmethod
    def process_exception(request, exception):
        if isinstance(exception, Custom410):
            return HttpResponseGone()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.api.function_trace import FunctionTraceWrapper
from newrelic.common.object_names import callable_name
from newrelic.common.object_wrapper import wrap_function_wrapper, function_wrapper
from newrelic.api.transaction import current_transaction
from newrelic.api.time_trace import notice_error


def _nr_wrap_handle_exception(wrapped, instance, args, kwargs):

    response = wrapped(*args, **kwargs)

    notice_error(status_code=response.status_code)

    return response


def outer_fn_wrapper(outer_fn, instance, args, kwargs):
    view_name = args[0]

    meta = getattr(instance, "_meta", None)

    if meta is None:
        group = 'Python/TastyPie/Api'
        name = instance.api_name
        callback = getattr(instance, 'top_level', None)
    elif meta.api_name is not None:
        group = 'Python/TastyPie/Api'
        name = '%s/%s/%s' % (meta.api_name, meta.resource_name, view_name)
        callback = getattr(instance, view_name, None)
    else:
        group = 'Python/TastyPie/Resource'
        name = '%s/%s' % (meta.resource_name, view_name)
        callback = getattr(instance, view_name, None)

    # Give preference to naming web transaction and trace node after
    # target callback, but fall back to abstract path if for some reason
    # we don't get a valid target callback.

    if callback is not None:
        name = callable_name(callback)
        group = None

    @function_wrapper
    def inner_fn_wrapper(inner_fn, instance, args, kwargs):
        transaction = current_transaction()

        if transaction is None or name is None:
            return inner_fn(*args, **kwargs)

        transaction.set_transaction_name(name, group, priority=5)

        # django's exception handling will record errors here
        return FunctionTraceWrapper(inner_fn, name=name, group=group)(*args, **kwargs)

    result = outer_fn(*args, **kwargs)

    return inner_fn_wrapper(result)


def instrument_tastypie_resources(module):
    wrap_function_wrapper(module, "Resource.wrap_view", outer_fn_wrapper)

    wrap_function_wrapper(module, 'Resource._handle_500', _nr_wrap_handle_exception)


def instrument_tastypie_api(module):
    wrap_function_wrapper(module, "Api.wrap_view", outer_fn_wrapper)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import os
import threading

import gearman

from newrelic.api.background_task import background_task
from testing_support.db_settings import gearman_settings

worker_thread = None
worker_event = threading.Event()

gm_client = None

GEARMAND_SETTINGS = gearman_settings()[0]
GEARMAND_HOST = GEARMAND_SETTINGS["host"]
GEARMAND_PORT = GEARMAND_SETTINGS["port"]

GEARMAND_ADDR = "%s:%s" % (GEARMAND_HOST, GEARMAND_PORT)


class GearmanWorker(gearman.GearmanWorker):
    def after_poll(self, any_activity):
        try:
            worker_event_set = worker_event.is_set()
        except TypeError:
            worker_event_set = worker_event.isSet()

        return not worker_event_set


def setup_module(module):
    global worker_thread

    gm_worker = GearmanWorker([GEARMAND_ADDR])

    def task_listener_reverse(gearman_worker, gearman_job):
        return "".join(reversed(gearman_job.data))

    def task_listener_exception(gearman_worker, gearman_job):
        raise RuntimeError("error")

    gm_worker.set_client_id("gearman-instrumentation-tests")
    gm_worker.register_task("reverse", task_listener_reverse)
    gm_worker.register_task("exception", task_listener_exception)

    def startup():
        gm_worker.work(poll_timeout=1.0)

    worker_thread = threading.Thread(target=startup)
    worker_thread.start()

    global gm_client

    gm_client = gearman.GearmanClient([GEARMAND_ADDR])


def teardown_module(module):
    worker_event.set()
    worker_thread.join()


@background_task()
def test_successful():
    completed_job_request = gm_client.submit_job("reverse", "data")
    assert completed_job_request.complete


@background_task()
def test_exception():
    completed_job_request = gm_client.submit_job("exception", "data")
    assert completed_job_request.complete



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import subprocess
import sys

import newrelic.agent

from newrelic.core.agent import agent_instance

from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_metric_payload import (
        validate_metric_payload)


_unscoped_metrics = [
        ('Supportability/api/FunctionTrace', 1),
        ('Supportability/api/current_transaction', 1),
        ('Supportability/api/callable_name', 1),
        ('Supportability/api/background_task', None),
]


@validate_transaction_metrics(
        'test_supportability_metrics:test_apis_in_transaction',
        custom_metrics=_unscoped_metrics,
        background_task=True)
@newrelic.agent.background_task()
def test_apis_in_transaction():
    transaction = newrelic.agent.current_transaction()
    name = newrelic.agent.callable_name(test_apis_in_transaction)
    with newrelic.agent.FunctionTrace(name):
        pass


_unscoped_metrics = [
        ('Supportability/api/global_settings', 2),
        ('Supportability/api/background_task', None),
]


@validate_transaction_metrics(
        'test_supportability_metrics:test_uses_api_twice',
        custom_metrics=_unscoped_metrics,
        background_task=True)
@newrelic.agent.background_task()
def test_uses_api_twice():
    newrelic.agent.global_settings()
    newrelic.agent.global_settings()


_unscoped_metrics = [
        ('Supportability/api/notice_error', 1),
        ('Supportability/api/background_task', None),
]

@validate_transaction_metrics(
        'test_supportability_metrics:test_notice_error',
        custom_metrics=_unscoped_metrics,
        background_task=True)
@newrelic.agent.background_task()
def test_notice_error():
    try:
        1 / 0
    except ZeroDivisionError:
        newrelic.agent.notice_error(sys.exc_info())


_unscoped_metrics = [
        ('Supportability/api/end_of_transaction', 1),
        ('Supportability/api/function_trace', None),
        ('Supportability/api/background_task', None),
]


@validate_transaction_metrics(
        'test_supportability_metrics:test_end_of_transaction',
        custom_metrics=_unscoped_metrics,
        background_task=True)
@newrelic.agent.background_task()
def test_end_of_transaction():
    # test that even if the transaction is ignored that we still create the
    # metric
    newrelic.agent.end_of_transaction()

    # in this case since there is no longer a transaction, the function_trace
    # metric should not be created here
    @newrelic.agent.function_trace()
    def _nothing():
        pass


@validate_metric_payload(metrics=[
        ('Supportability/Python/Uninstrumented', None),
])
def test_uninstrumented_none():
    # tests a bug that returned "TypeError: 'NoneType' object is not iterable"
    app_name = 'Python Agent Test (uninstrumented 3)'
    agent = agent_instance()
    agent.activate_application(app_name)
    application = agent._applications.get(app_name)
    application.harvest()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import redis
from testing_support.db_settings import redis_settings

DB_SETTINGS = redis_settings()[0]

redis_client = redis.Redis(host=DB_SETTINGS["host"], port=DB_SETTINGS["port"], db=0)
strict_redis_client = redis.StrictRedis(host=DB_SETTINGS["host"], port=DB_SETTINGS["port"], db=0)


IGNORED_METHODS = {
    "MODULE_CALLBACKS",
    "MODULE_VERSION",
    "NAME",
    "add_edge",
    "add_node",
    "append_bucket_size",
    "append_capacity",
    "append_error",
    "append_expansion",
    "append_items_and_increments",
    "append_items",
    "append_max_iterations",
    "append_no_create",
    "append_no_scale",
    "append_values_and_weights",
    "append_weights",
    "auto_close_connection_pool",
    "batch_indexer",
    "BatchIndexer",
    "bulk",
    "call_procedure",
    "client_no_touch",
    "client_tracking_off",
    "client_tracking_on",
    "client",
    "close",
    "commandmixin",
    "connection_pool",
    "connection",
    "debug_segfault",
    "edges",
    "execute_command",
    "flush",
    "from_pool",
    "from_url",
    "get_connection_kwargs",
    "get_encoder",
    "get_label",
    "get_params_args",
    "get_property",
    "get_relation",
    "get_retry",
    "index_name",
    "labels",
    "list_keys",
    "load_document",
    "load_external_module",
    "lock",
    "name",
    "nodes",
    "parse_response",
    "pipeline",
    "property_keys",
    "register_script",
    "relationship_types",
    "response_callbacks",
    "RESPONSE_CALLBACKS",
    "sentinel",
    "set_file",
    "set_path",
    "set_response_callback",
    "set_retry",
    "transaction",
    "version",
}

REDIS_MODULES = {
    "bf",
    "cf",
    "cms",
    "ft",
    "graph",
    "json",
    "tdigest",
    "topk",
    "ts",
}

IGNORED_METHODS |= REDIS_MODULES


@pytest.mark.parametrize("client", (redis_client, strict_redis_client))
def test_uninstrumented_methods(client):
    methods = {m for m in dir(client) if not m[0] == "_"}
    is_wrapped = lambda m: hasattr(getattr(client, m), "__wrapped__")
    uninstrumented = {m for m in methods - IGNORED_METHODS if not is_wrapped(m)}

    for module in REDIS_MODULES:
        if hasattr(client, module):
            module_client = getattr(client, module)()
            module_methods = {m for m in dir(module_client) if not m[0] == "_"}
            is_wrapped = lambda m: hasattr(getattr(module_client, m), "__wrapped__")
            uninstrumented |= {m for m in module_methods - IGNORED_METHODS if not is_wrapped(m)}

    assert not uninstrumented, "Uninstrumented methods: %s" % sorted(uninstrumented)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import psycopg2cffi
import psycopg2cffi.extensions
import psycopg2cffi.extras
import pytest

from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_database_node import (
    validate_database_node,
)
from testing_support.validators.validate_transaction_slow_sql_count import \
        validate_transaction_slow_sql_count
from newrelic.api.background_task import background_task
from newrelic.core.database_utils import SQLConnections

from testing_support.db_settings import postgresql_settings
DB_SETTINGS = postgresql_settings()[0]


def _exercise_db(cursor_kwargs=None):
    cursor_kwargs = cursor_kwargs or {}

    connection = psycopg2cffi.connect(
            database=DB_SETTINGS['name'], user=DB_SETTINGS['user'],
            password=DB_SETTINGS['password'], host=DB_SETTINGS['host'],
            port=DB_SETTINGS['port'])

    try:
        cursor = connection.cursor(**cursor_kwargs)

        cursor.execute("""SELECT setting from pg_settings where name=%s""",
                ('server_version',))
    finally:
        connection.close()


# Tests


def explain_plan_is_not_none(node):
    with SQLConnections() as connections:
        explain_plan = node.explain_plan(connections)

    assert explain_plan is not None


@override_application_settings({
    'transaction_tracer.explain_threshold': 0.0,
    'transaction_tracer.record_sql': 'raw',
})
@validate_database_node(explain_plan_is_not_none)
@validate_transaction_slow_sql_count(1)
@background_task(name='test_explain_plan_named_cursors')
@pytest.mark.parametrize('withhold', (True, False))
@pytest.mark.parametrize('scrollable', (True, False))
def test_explain_plan_named_cursors(withhold, scrollable):
    cursor_kwargs = {
        'name': 'test_explain_plan_named_cursors',
    }

    if withhold:
        cursor_kwargs['withhold'] = withhold

    if scrollable:
        cursor_kwargs['scrollable'] = scrollable

    _exercise_db(cursor_kwargs=cursor_kwargs)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncpg
import pytest
from testing_support.db_settings import postgresql_settings
from testing_support.fixtures import override_application_settings
from testing_support.util import instance_hostname
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task
from newrelic.common.package_version_utils import get_package_version_tuple

DB_MULTIPLE_SETTINGS = postgresql_settings()

ASYNCPG_VERSION = get_package_version_tuple("asyncpg")

if ASYNCPG_VERSION < (0, 11):
    CONNECT_METRICS = []

    # Only 1 statement will be recorded since connect is not captured
    STATEMENT_COUNT = 1

    # 2 DBs are queried
    TOTAL_COUNT = STATEMENT_COUNT * 2
else:
    CONNECT_METRICS = [("Datastore/operation/Postgres/connect", 2)]

    # Two statements are executed per DB - connect / select
    STATEMENT_COUNT = 2

    # 2 DBs are queried
    TOTAL_COUNT = STATEMENT_COUNT * 2


# Settings

_enable_instance_settings = {
    "datastore_tracer.instance_reporting.enabled": True,
}
_disable_instance_settings = {
    "datastore_tracer.instance_reporting.enabled": False,
}


# Metrics

_base_scoped_metrics = CONNECT_METRICS + [
    ("Datastore/statement/Postgres/pg_settings/select", 2),
]

_base_rollup_metrics = [
    ("Datastore/all", TOTAL_COUNT),
    ("Datastore/allOther", TOTAL_COUNT),
    ("Datastore/Postgres/all", TOTAL_COUNT),
    ("Datastore/Postgres/allOther", TOTAL_COUNT),
    ("Datastore/statement/Postgres/pg_settings/select", 2),
]

_enable_scoped_metrics = list(_base_scoped_metrics)
_enable_rollup_metrics = list(_base_rollup_metrics)

_disable_scoped_metrics = list(_base_scoped_metrics)
_disable_rollup_metrics = list(_base_rollup_metrics)

if len(DB_MULTIPLE_SETTINGS) > 1:
    _postgresql_1 = DB_MULTIPLE_SETTINGS[0]
    _host_1 = instance_hostname(_postgresql_1["host"])
    _port_1 = _postgresql_1["port"]

    _postgresql_2 = DB_MULTIPLE_SETTINGS[1]
    _host_2 = instance_hostname(_postgresql_2["host"])
    _port_2 = _postgresql_2["port"]

    _instance_metric_name_1 = "Datastore/instance/Postgres/%s/%s" % (_host_1, _port_1)
    _instance_metric_name_2 = "Datastore/instance/Postgres/%s/%s" % (_host_2, _port_2)

    _enable_rollup_metrics.extend(
        [
            (_instance_metric_name_1, STATEMENT_COUNT),
            (_instance_metric_name_2, STATEMENT_COUNT),
        ]
    )
    _disable_rollup_metrics.extend([(_instance_metric_name_1, None), (_instance_metric_name_2, None)])


# Query


async def _exercise_db():
    postgresql1 = DB_MULTIPLE_SETTINGS[0]
    postgresql2 = DB_MULTIPLE_SETTINGS[1]

    connection = await asyncpg.connect(
        user=postgresql1["user"],
        password=postgresql1["password"],
        database=postgresql1["name"],
        host=postgresql1["host"],
        port=postgresql1["port"],
    )
    try:
        await connection.execute("SELECT setting from pg_settings where name='server_version'")
    finally:
        await connection.close()

    connection = await asyncpg.connect(
        user=postgresql2["user"],
        password=postgresql2["password"],
        database=postgresql2["name"],
        host=postgresql2["host"],
        port=postgresql2["port"],
    )
    try:
        await connection.execute("SELECT setting from pg_settings where name='server_version'")
    finally:
        await connection.close()


# Tests


@pytest.mark.skipif(
    len(DB_MULTIPLE_SETTINGS) < 2,
    reason="Test environment not configured with multiple databases.",
)
@override_application_settings(_enable_instance_settings)
@validate_transaction_metrics(
    "test_multiple_dbs:test_multiple_databases_enable_instance",
    scoped_metrics=_enable_scoped_metrics,
    rollup_metrics=_enable_rollup_metrics,
    background_task=True,
)
@background_task()
def test_multiple_databases_enable_instance(event_loop):
    assert ASYNCPG_VERSION is not None
    event_loop.run_until_complete(_exercise_db())


@pytest.mark.skipif(
    len(DB_MULTIPLE_SETTINGS) < 2,
    reason="Test environment not configured with multiple databases.",
)
@override_application_settings(_disable_instance_settings)
@validate_transaction_metrics(
    "test_multiple_dbs:test_multiple_databases_disable_instance",
    scoped_metrics=_disable_scoped_metrics,
    rollup_metrics=_disable_scoped_metrics,
    background_task=True,
)
@background_task()
def test_multiple_databases_disable_instance(event_loop):
    assert ASYNCPG_VERSION is not None
    event_loop.run_until_complete(_exercise_db())



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64

import pytest
import webtest
from bottle import __version__ as version
from testing_support.fixtures import (
    override_application_settings,
    override_ignore_status_codes,
)
from testing_support.validators.validate_code_level_metrics import (
    validate_code_level_metrics,
)
from testing_support.validators.validate_transaction_errors import (
    validate_transaction_errors,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.common.package_version_utils import get_package_version_tuple
from newrelic.packages import six

version = list(get_package_version_tuple("bottle"))

if len(version) == 2:
    version.append(0)

version = tuple(version)
assert version > (0, 1), "version information not found"

requires_auth_basic = pytest.mark.skipif(version < (0, 9, 0), reason="Bottle only added auth_basic in 0.9.0.")
requires_plugins = pytest.mark.skipif(version < (0, 9, 0), reason="Bottle only added auth_basic in 0.9.0.")

_test_application_index_scoped_metrics = [
    ("Python/WSGI/Application", 1),
    ("Python/WSGI/Response", 1),
    ("Python/WSGI/Finalize", 1),
    ("Function/_target_application:index_page", 1),
]

if version >= (0, 9, 0):
    _test_application_index_scoped_metrics.extend([("Function/bottle:Bottle.wsgi", 1)])
else:
    _test_application_index_scoped_metrics.extend([("Function/bottle:Bottle.__call__", 1)])

_test_application_index_custom_metrics = [("Python/Framework/Bottle/%s.%s.%s" % version, 1)]


@validate_code_level_metrics("_target_application", "index_page")
@validate_transaction_errors(errors=[])
@validate_transaction_metrics(
    "_target_application:index_page",
    scoped_metrics=_test_application_index_scoped_metrics,
    custom_metrics=_test_application_index_custom_metrics,
)
def test_application_index(target_application):
    response = target_application.get("/index")
    response.mustcontain("INDEX RESPONSE")


_test_application_error_scoped_metrics = [
    ("Python/WSGI/Application", 1),
    ("Python/WSGI/Response", 1),
    ("Python/WSGI/Finalize", 1),
    ("Function/_target_application:error_page", 1),
]

if version >= (0, 9, 0):
    _test_application_error_scoped_metrics.extend([("Function/bottle:Bottle.wsgi", 1)])
else:
    _test_application_error_scoped_metrics.extend([("Function/bottle:Bottle.__call__", 1)])

_test_application_error_custom_metrics = [("Python/Framework/Bottle/%s.%s.%s" % version, 1)]

if six.PY3:
    _test_application_error_errors = ["builtins:RuntimeError"]
else:
    _test_application_error_errors = ["exceptions:RuntimeError"]


@validate_code_level_metrics("_target_application", "error_page")
@validate_transaction_errors(errors=_test_application_error_errors)
@validate_transaction_metrics(
    "_target_application:error_page",
    scoped_metrics=_test_application_error_scoped_metrics,
    custom_metrics=_test_application_error_custom_metrics,
)
def test_application_error(target_application):
    response = target_application.get("/error", status=500, expect_errors=True)


_test_application_not_found_scoped_metrics = [
    ("Python/WSGI/Application", 1),
    ("Python/WSGI/Response", 1),
    ("Python/WSGI/Finalize", 1),
    ("Function/_target_application:error404_page", 1),
]

if version >= (0, 9, 0):
    _test_application_not_found_scoped_metrics.extend([("Function/bottle:Bottle.wsgi", 1)])
else:
    _test_application_not_found_scoped_metrics.extend([("Function/bottle:Bottle.__call__", 1)])

_test_application_not_found_custom_metrics = [("Python/Framework/Bottle/%s.%s.%s" % version, 1)]


@validate_code_level_metrics("_target_application", "error404_page")
@validate_transaction_errors(errors=[])
@validate_transaction_metrics(
    "_target_application:error404_page",
    scoped_metrics=_test_application_not_found_scoped_metrics,
    custom_metrics=_test_application_not_found_custom_metrics,
)
def test_application_not_found(target_application):
    response = target_application.get("/missing", status=404)
    response.mustcontain("NOT FOUND")


_test_application_auth_basic_fail_scoped_metrics = [
    ("Python/WSGI/Application", 1),
    ("Python/WSGI/Response", 1),
    ("Python/WSGI/Finalize", 1),
    ("Function/_target_application:auth_basic_page", 1),
]

if version >= (0, 9, 0):
    _test_application_auth_basic_fail_scoped_metrics.extend([("Function/bottle:Bottle.wsgi", 1)])
else:
    _test_application_auth_basic_fail_scoped_metrics.extend([("Function/bottle:Bottle.__call__", 1)])

_test_application_auth_basic_fail_custom_metrics = [("Python/Framework/Bottle/%s.%s.%s" % version, 1)]


@requires_auth_basic
@validate_code_level_metrics("_target_application", "auth_basic_page")
@validate_transaction_errors(errors=[])
@validate_transaction_metrics(
    "_target_application:auth_basic_page",
    scoped_metrics=_test_application_auth_basic_fail_scoped_metrics,
    custom_metrics=_test_application_auth_basic_fail_custom_metrics,
)
def test_application_auth_basic_fail(target_application):
    response = target_application.get("/auth", status=401)


_test_application_auth_basic_okay_scoped_metrics = [
    ("Python/WSGI/Application", 1),
    ("Python/WSGI/Response", 1),
    ("Python/WSGI/Finalize", 1),
    ("Function/_target_application:auth_basic_page", 1),
]

if version >= (0, 9, 0):
    _test_application_auth_basic_okay_scoped_metrics.extend([("Function/bottle:Bottle.wsgi", 1)])
else:
    _test_application_auth_basic_okay_scoped_metrics.extend([("Function/bottle:Bottle.__call__", 1)])

_test_application_auth_basic_okay_custom_metrics = [("Python/Framework/Bottle/%s.%s.%s" % version, 1)]


@requires_auth_basic
@validate_code_level_metrics("_target_application", "auth_basic_page")
@validate_transaction_errors(errors=[])
@validate_transaction_metrics(
    "_target_application:auth_basic_page",
    scoped_metrics=_test_application_auth_basic_okay_scoped_metrics,
    custom_metrics=_test_application_auth_basic_okay_custom_metrics,
)
def test_application_auth_basic_okay(target_application):
    authorization_value = base64.b64encode(b"user:password")
    if six.PY3:
        authorization_value = authorization_value.decode("Latin-1")
    environ = {"HTTP_AUTHORIZATION": "Basic " + authorization_value}
    response = target_application.get("/auth", extra_environ=environ)
    response.mustcontain("AUTH OKAY")


_test_application_plugin_error_scoped_metrics = [
    ("Python/WSGI/Application", 1),
    ("Python/WSGI/Response", 1),
    ("Python/WSGI/Finalize", 1),
    ("Function/_target_application:plugin_error_page", 1),
]

if version >= (0, 9, 0):
    _test_application_plugin_error_scoped_metrics.extend([("Function/bottle:Bottle.wsgi", 1)])
else:
    _test_application_plugin_error_scoped_metrics.extend([("Function/bottle:Bottle.__call__", 1)])

_test_application_plugin_error_custom_metrics = [("Python/Framework/Bottle/%s.%s.%s" % version, 1)]


@requires_plugins
@validate_code_level_metrics("_target_application", "plugin_error_page")
@validate_transaction_errors(errors=[])
@validate_transaction_metrics(
    "_target_application:plugin_error_page",
    scoped_metrics=_test_application_plugin_error_scoped_metrics,
    custom_metrics=_test_application_plugin_error_custom_metrics,
)
@override_ignore_status_codes([403])
def test_application_plugin_error_ignore(target_application):
    response = target_application.get("/plugin_error", status=403, expect_errors=True)


@requires_plugins
@validate_code_level_metrics("_target_application", "plugin_error_page")
@validate_transaction_errors(errors=["bottle:HTTPError"])
@validate_transaction_metrics(
    "_target_application:plugin_error_page",
    scoped_metrics=_test_application_plugin_error_scoped_metrics,
    custom_metrics=_test_application_plugin_error_custom_metrics,
)
def test_application_plugin_error_capture(target_application):
    import newrelic.agent

    response = target_application.get("/plugin_error", status=403, expect_errors=True)


_test_html_insertion_settings = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": True,
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_html_insertion_settings)
def test_html_insertion(target_application):
    response = target_application.get("/html_insertion")

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # header added by the agent.

    response.mustcontain("NREUM HEADER", "NREUM.info")



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

from newrelic.common.encoding_utils import json_encode
from newrelic.common.object_wrapper import transient_function_wrapper
from newrelic.packages import six


def validate_error_trace_collector_json():
    @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.record_transaction")
    def _validate_error_trace_collector_json(wrapped, instance, args, kwargs):
        try:
            result = wrapped(*args, **kwargs)
        except:
            raise
        else:
            errors = instance.error_data()

            # recreate what happens right before data is sent to the collector
            # in data_collector.py via ApplicationSession.send_errors
            agent_run_id = 666
            payload = (agent_run_id, errors)
            collector_json = json_encode(payload)

            decoded_json = json.loads(collector_json)

            assert decoded_json[0] == agent_run_id
            err = decoded_json[1][0]
            assert len(err) == 5
            assert isinstance(err[0], (int, float))
            assert isinstance(err[1], six.string_types)  # path
            assert isinstance(err[2], six.string_types)  # error message
            assert isinstance(err[3], six.string_types)  # exception name
            parameters = err[4]

            parameter_fields = ["userAttributes", "stack_trace", "agentAttributes", "intrinsics"]

            for field in parameter_fields:
                assert field in parameters

            assert "request_uri" not in parameters

        return result

    return _validate_error_trace_collector_json



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.api.datastore_trace import DatastoreTraceWrapper
from newrelic.api.transaction import current_transaction
from newrelic.common.object_wrapper import wrap_function_wrapper
from newrelic.packages import six

# An index name can be a string, None or a sequence. In the case of None
# an empty string or '*', it is the same as using '_all'. When a string
# it can also be a comma separated list of index names. A sequence
# obviously can also be more than one index name. Where we are certain
# there is only a single index name we use it, otherwise we use 'other'.


def _index_name(index):
    if not index or index == "*":
        return "_all"
    if not isinstance(index, six.string_types) or "," in index:
        return "other"
    return index


def _extract_kwargs_index(*args, **kwargs):
    return _index_name(kwargs.get("index"))


def _extract_args_index(index=None, *args, **kwargs):
    return _index_name(index)


def _extract_args_metric_index(metric=None, index=None, *args, **kwargs):
    return _index_name(index)


_elasticsearch_client_methods = (
    ("bulk", None),
    ("bulk_index", _extract_args_index),
    ("close_index", None),
    ("cluster_state", _extract_args_metric_index),
    ("count", _extract_kwargs_index),
    ("create_index", _extract_args_index),
    ("delete", _extract_args_index),
    ("delete_all", _extract_args_index),
    ("delete_all_indexes", None),
    ("delete_by_query", _extract_args_index),
    ("delete_index", _extract_args_index),
    ("flush", _extract_args_index),
    ("gateway_snapshot", _extract_args_index),
    ("get", _extract_args_index),
    ("get_aliases", _extract_args_index),
    ("get_mapping", _extract_args_index),
    ("get_settings", _extract_args_index),
    ("health", _extract_args_index),
    ("index", _extract_args_index),
    ("more_like_this", _extract_args_index),
    ("multi_get", None),
    ("open_index", _extract_args_index),
    ("optimize", _extract_args_index),
    ("percolate", _extract_args_index),
    ("put_mapping", _extract_args_index),
    ("refresh", _extract_args_index),
    ("search", _extract_kwargs_index),
    ("send_request", None),
    ("status", _extract_args_index),
    ("update", _extract_args_index),
    ("update_aliases", None),
    ("update_all_settings", None),
    ("update_settings", _extract_args_index),
)


def wrap_elasticsearch_client_method(module, name, arg_extractor):
    def _nr_wrapper_ElasticSearch_method_(wrapped, instance, args, kwargs):
        transaction = current_transaction()

        if transaction is None:
            return wrapped(*args, **kwargs)

        # When arg_extractor is None, it means there is no target field
        # associated with this method. Hence this method will only
        # create an operation metric and no statement metric. This is
        # handled by setting the target to None when calling the
        # DatastoreTraceWrapper.

        if arg_extractor is None:
            index = None
        else:
            index = arg_extractor(*args, **kwargs)

        return DatastoreTraceWrapper(wrapped, product="Elasticsearch", target=index, operation=name)(*args, **kwargs)

    if hasattr(module.ElasticSearch, name):
        wrap_function_wrapper(module, "ElasticSearch.%s" % name, _nr_wrapper_ElasticSearch_method_)


def instrument_pyelasticsearch_client(module):
    for name, arg_extractor in _elasticsearch_client_methods:
        wrap_elasticsearch_client_method(module, name, arg_extractor)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import pytest

from newrelic.config import _process_labels_setting, _map_labels
from newrelic.core.config import global_settings

from testing_support.fixtures import override_application_settings

CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
FIXTURE = os.path.join(CURRENT_DIR, 'fixtures', 'labels.json')

def _load_tests():
    with open(FIXTURE, 'r') as fh:
        js = fh.read()
    return json.loads(js)

def _parametrize_test(test):

    # pytest.mark.parametrize expects each test to be a tuple

    return tuple([test['name'], test['labelString'],
            test['warning'], test['expected']])

_labels_tests = [_parametrize_test(t) for t in _load_tests()]

@pytest.mark.parametrize('name,labelString,warning,expected', _labels_tests)
def test_labels(name, labelString, warning, expected):

    parsed_labels = _map_labels(labelString)
    _process_labels_setting(parsed_labels)

    settings = global_settings()

    sorted_labels = sorted(settings.labels, key=lambda x: x['label_type'])
    sorted_expected = sorted(expected, key=lambda x: x['label_type'])

    assert sorted_labels == sorted_expected
